"""Configuration to obtain registry classes"""
import importlib
import logging

from . import example_filetype_format

logger = logging.getLogger(__name__)

# BASE_CLASS = example_filetype_format.FileTypeFormat


def make_format_registry_dict(cls_list):
    """Use an object's _fileType attribute to make a class lookup dictionary.

    Args:
        cls_list: A list of Python classes.

    Returns:
        A dictionary mapping the class._fileType to the class.

    """
    return {cls._fileType: cls for cls in cls_list}


def get_subclasses(cls):
    """Gets subclasses of modules and classes"""
    for subclass in cls.__subclasses__():
        yield from get_subclasses(subclass)
        yield subclass


def find_subclasses(package_names, base_class):
    """Finds subclasses of a specified base class
    from a list of package names.

    Args:
        package_names: A list of Python package names as strings.
        base_class: A base class to use to search for subclasses

    Returns:
        A list of subclasses that extend from the base class

    """
    matching_classes = []
    for package_name in package_names:
        importlib.import_module(package_name)

    for cls in get_subclasses(base_class):
        logger.debug("checking {cls}.".format(cls=cls))
        cls_module_name = cls.__module__
        cls_pkg = cls_module_name.split(".")[0]
        if cls_pkg in package_names:
            matching_classes.append(cls)
    return matching_classes


def collect_format_types(package_names):
    """Finds subclasses of the example_filetype_format.FileTypeFormat from a
    list of package names.

    Args:
        package_names: A list of Python package names as strings.

    Returns:
        A list of classes that are in the named packages and subclasses of
        example_filetype_format.FileTypeFormat.

    """
    file_format_list = find_subclasses(
        package_names, example_filetype_format.FileTypeFormat
    )
    file_format_dict = make_format_registry_dict(file_format_list)
    return file_format_dict


# PROCESS_FILES_LIST = [x for x in get_subclasses(BASE_CLASS)]
# PROCESS_FILES = make_format_registry_dict(cls_list=PROCESS_FILES_LIST)



import argparse
import datetime
import logging
import os
import subprocess
import synapseclient

# import time

from genie import (
    create_case_lists,
    dashboard_table_updater,
    database_to_staging,
    process_functions,
)

logger = logging.getLogger(__name__)

PWD = os.path.dirname(os.path.abspath(__file__))


def generate_dashboard_html(
    genie_version, staging=False, genie_user=None, genie_pass=None
):
    """Generates dashboard html writeout that gets uploaded to the
    release folder

    Args:
        syn: Synapse connection
        genie_version: GENIE release
        staging: Use staging files. Default is False
        genie_user: GENIE synapse username
        genie_pass: GENIE synapse password

    """
    markdown_render_cmd = [
        "Rscript",
        os.path.join(PWD, "../R/dashboard_markdown_generator.R"),
        genie_version,
        "--template_path",
        os.path.join(PWD, "../templates/dashboardTemplate.Rmd"),
    ]

    if genie_user is not None and genie_pass is not None:
        markdown_render_cmd.extend(["--syn_user", genie_user, "--syn_pass", genie_pass])
    if staging:
        markdown_render_cmd.append("--staging")
    subprocess.check_call(markdown_render_cmd)


def generate_data_guide(
    genie_version,
    oncotree_version=None,
    database_mapping=None,
    genie_user=None,
    genie_pass=None,
):
    """Generates the GENIE data guide"""

    template_path = os.path.join(PWD, "../templates/data_guide_template.Rnw")
    with open(template_path, "r") as template_file:
        template_str = template_file.read()

    replacements = {
        "{{release}}": genie_version,
        "{{database_synid}}": database_mapping,
        "{{oncotree}}": oncotree_version.replace("_", "\\_"),
        "{{username}}": genie_user,
        "{{password}}": genie_pass,
        "{{genie_banner}}": os.path.join(PWD, "../genie_banner.png"),
    }

    for search in replacements:
        replacement = replacements[search]
        # If no replacement value is passed in, don't replace
        if replacement is not None:
            template_str = template_str.replace(search, replacement)

    with open(os.path.join(PWD, "data_guide.Rnw"), "w") as data_guide_file:
        data_guide_file.write(template_str)

    subprocess.check_call(
        ["R", "CMD", "Sweave", "--pdf", os.path.join(PWD, "data_guide.Rnw")]
    )
    return "data_guide.pdf"


def main(
    genie_version,
    processing_date,
    cbioportal_path,
    oncotree_link=None,
    consortium_release_cutoff=184,
    pemfile=None,
    test=False,
    staging=False,
    debug=False,
    skip_mutationsincis=False,
):
    """
    - Does parameter checks
    - Updates process tracking start
    - initiates database to staging
    - create case lists
    - revise meta files
    - run cBioPortal validation
    - create link versions
    - update process tracking end
    - Create dashboard tables and plots

    Args:
        genie_version: GENIE version,
        processing_date: processing date
        cbioportal_path: Path to cbioportal validator
        oncotree_link: Link to oncotree codes
        consortium_release_cutoff: release cut off value in days
        pemfile: Path to private key file
        test: Test flag, uses test databases
        staging: Staging flag, uses staging databases
        debug:  Synapse debug flag
        skip_mutationsincis: Skip mutation in cis filter
    """
    syn = process_functions.synLogin(pemfile, debug=debug)
    genie_user = os.environ.get("GENIE_USER")
    if pemfile is not None:
        genie_pass = process_functions.get_password(pemfile)
    else:
        genie_pass = None

    if test:
        databaseSynIdMappingId = "syn11600968"
        genie_version = "TESTING"
    elif staging:
        skip_mutationsincis = True
        databaseSynIdMappingId = "syn12094210"
    else:
        databaseSynIdMappingId = "syn10967259"
    # Database/folder syn id mapping
    databaseSynIdMapping = syn.tableQuery(
        "select * from {}".format(databaseSynIdMappingId)
    )
    databaseSynIdMappingDf = databaseSynIdMapping.asDataFrame()
    # databaseSynIdMappingDf.index = databaseSynIdMappingDf.Database
    # del databaseSynIdMappingDf['Database']
    # databaseSynIdMappingDf.to_dict()

    if oncotree_link is None:
        oncoLink = databaseSynIdMappingDf["Id"][
            databaseSynIdMappingDf["Database"] == "oncotreeLink"
        ].values[0]
        oncoLinkEnt = syn.get(oncoLink)
        oncotree_link = oncoLinkEnt.externalURL

    # Check if you can connect to oncotree link,
    # if not then don't run validation / processing
    process_functions.checkUrl(oncotree_link)

    cbioValidatorPath = os.path.join(
        cbioportal_path, "core/src/main/scripts/importer/validateData.py"
    )
    assert os.path.exists(cbioValidatorPath), "Please specify correct cbioportalPath"
    syn.table_query_timeout = 50000

    consortiumSynId = databaseSynIdMappingDf["Id"][
        databaseSynIdMappingDf["Database"] == "consortium"
    ].values[0]
    processTrackerSynId = databaseSynIdMappingDf["Id"][
        databaseSynIdMappingDf["Database"] == "processTracker"
    ].values[0]
    # get syn id of case list folder in consortium release
    # caseListSynId = findCaseListId(syn, consortiumSynId)
    caseListSynId, _ = database_to_staging.search_and_create_folder(
        syn, consortiumSynId, "case_lists"
    )

    if not staging:
        database_to_staging.update_process_trackingdf(
            syn, processTrackerSynId, "SAGE", "dbToStage", start=True
        )

    centerMappingSynId = databaseSynIdMappingDf["Id"][
        databaseSynIdMappingDf["Database"] == "centerMapping"
    ].values[0]
    # Only release files where release is true
    center_mapping = syn.tableQuery(
        "SELECT * FROM {} where release is true".format(centerMappingSynId)
    )
    center_mappingdf = center_mapping.asDataFrame()
    processingDate = datetime.datetime.strptime(processing_date, "%b-%Y")

    logger.info("STAGING TO CONSORTIUM")
    genePanelEntities = database_to_staging.stagingToCbio(
        syn,
        processingDate,
        genie_version,
        center_mappingdf,
        databaseSynIdMappingDf,
        oncotree_url=oncotree_link,
        consortiumReleaseCutOff=consortium_release_cutoff,
        current_release_staging=staging,
        skipMutationsInCis=skip_mutationsincis,
        test=test,
        genie_user=genie_user,
        genie_pass=genie_pass,
    )

    # Create case lists files
    logger.info("CREATE CASE LIST FILES")
    # Remove old caselists first
    if not os.path.exists(database_to_staging.CASE_LIST_PATH):
        os.mkdir(database_to_staging.CASE_LIST_PATH)
    caselists = os.listdir(database_to_staging.CASE_LIST_PATH)
    for caselist in caselists:
        os.remove(os.path.join(database_to_staging.CASE_LIST_PATH, caselist))
    clinical_path = os.path.join(
        database_to_staging.GENIE_RELEASE_DIR,
        "data_clinical_{}.txt".format(genie_version),
    )
    assay_information_path = os.path.join(
        database_to_staging.GENIE_RELEASE_DIR,
        "assay_information_{}.txt".format(genie_version),
    )
    create_case_lists.main(
        clinical_path,
        assay_information_path,
        database_to_staging.CASE_LIST_PATH,
        "genie_private",
    )
    caseListFiles = os.listdir(database_to_staging.CASE_LIST_PATH)
    caseListEntities = []
    for casePath in caseListFiles:
        casePath = os.path.join(database_to_staging.CASE_LIST_PATH, casePath)
        caseListEntities.append(
            database_to_staging.store_file(
                syn, casePath, parent=caseListSynId, genieVersion=genie_version
            )
        )

    logger.info("REMOVING UNNECESSARY FILES")
    genie_files = os.listdir(database_to_staging.GENIE_RELEASE_DIR)
    for genie_file in genie_files:
        if (
            genie_version not in genie_file
            and "meta" not in genie_file
            and "case_lists" not in genie_file
        ):
            os.remove(os.path.join(database_to_staging.GENIE_RELEASE_DIR, genie_file))
    os.remove(clinical_path)

    logger.info("REVISE METADATA FILES")
    database_to_staging.revise_metadata_files(
        syn, staging, consortiumSynId, genie_version
    )

    logger.info("CBIO VALIDATION")

    # Must be exit 0 because the validator sometimes fails,
    # but we still want to capture the output

    command = [
        cbioValidatorPath,
        "-s",
        database_to_staging.GENIE_RELEASE_DIR,
        "-n",
        "; exit 0",
    ]
    cbioOutput = subprocess.check_output(" ".join(command), shell=True)
    logger.info(cbioOutput.decode("utf-8"))

    cbio_validator_log = f"cbioValidatorLogsConsortium_{genie_version}.txt"
    if not test and not staging:
        log_folder_synid = databaseSynIdMappingDf["Id"][
            databaseSynIdMappingDf["Database"] == "logs"
        ].values[0]
        with open(cbio_validator_log, "w") as cbio_log:
            cbio_log.write(cbioOutput.decode("utf-8"))
        syn.store(synapseclient.File(cbio_validator_log, parentId=log_folder_synid))
        os.remove(cbio_validator_log)
    logger.info("REMOVING OLD FILES")

    process_functions.rmFiles(database_to_staging.CASE_LIST_PATH)
    private_cna_meta_path = os.path.join(
        database_to_staging.GENIE_RELEASE_DIR, "genie_private_meta_cna_hg19_seg.txt"
    )
    if os.path.exists(private_cna_meta_path):
        os.unlink(private_cna_meta_path)

    logger.info("CREATING LINK VERSION")
    # Returns release and case list folder
    folders = database_to_staging.create_link_version(
        syn, genie_version, caseListEntities, genePanelEntities, databaseSynIdMappingDf
    )

    if not staging:
        database_to_staging.update_process_trackingdf(
            syn, processTrackerSynId, "SAGE", "dbToStage", start=False
        )

    if not test:
        logger.info("DASHBOARD UPDATE")
        dashboard_table_updater.run_dashboard(
            syn, databaseSynIdMappingDf, genie_version, staging=staging
        )
        generate_dashboard_html(
            genie_version, staging=staging, genie_user=genie_user, genie_pass=genie_pass
        )
        logger.info("DASHBOARD UPDATE COMPLETE")
        logger.info("AUTO GENERATE DATA GUIDE")

    oncotree_version = oncotree_link.split("=")[1]
    data_guide_pdf = generate_data_guide(
        genie_version,
        oncotree_version=oncotree_version,
        database_mapping=databaseSynIdMappingId,
        genie_user=genie_user,
        genie_pass=genie_pass,
    )
    database_to_staging.store_file(
        syn,
        data_guide_pdf,
        genieVersion=genie_version,
        parent=folders["release_folder"],
    )
    logger.info("COMPLETED DATABASE TO STAGING")


if __name__ == "__main__":
    parser = argparse.ArgumentParser(description="Release GENIE consortium files")

    parser.add_argument(
        "processingDate",
        type=str,
        metavar="Jan-2017",
        help="The processing date of GENIE in Month-Year format" " (ie. Apr-2017)",
    )

    parser.add_argument(
        "cbioportalPath",
        type=str,
        metavar="/path/to/cbioportal",
        help="Make sure you clone the cbioportal github: "
        "git clone https://github.com/cBioPortal/cbioportal.git",
    )

    parser.add_argument(
        "genieVersion", type=str, metavar="1.0.1", help="GENIE release version"
    )

    parser.add_argument("--oncotree_link", type=str, help="Link to oncotree code")

    parser.add_argument(
        "--consortiumReleaseCutOff",
        type=int,
        metavar=184,
        default=184,
        help="Consortium release cut off time in days",
    )

    parser.add_argument(
        "--skipMutationsInCis",
        action="store_true",
        help="Skip running mutation in cis script",
    )

    parser.add_argument("--pemFile", type=str, help="Path to PEM file (genie.pem)")

    parser.add_argument("--debug", action="store_true", help="Synapse debug feature")

    test_group = parser.add_mutually_exclusive_group()
    test_group.add_argument("--test", action="store_true", help="Run test")

    test_group.add_argument(
        "--staging", action="store_true", help="Store into staging folder"
    )

    args = parser.parse_args()

    main(
        genie_version=args.genieVersion,
        processing_date=args.processingDate,
        cbioportal_path=args.cbioportalPath,
        oncotree_link=args.oncotree_link,
        consortium_release_cutoff=args.consortiumReleaseCutOff,
        pemfile=args.pemFile,
        test=args.test,
        staging=args.staging,
        debug=args.debug,
        skip_mutationsincis=args.skipMutationsInCis,
    )



import logging
import os

import pandas as pd

from genie.example_filetype_format import FileTypeFormat
from genie import process_functions

logger = logging.getLogger(__name__)


def contains_whitespace(row):
    """Gets the total number of whitespaces from each column of a row"""
    return sum([" " in i for i in row if isinstance(i, str)])


class vcf(FileTypeFormat):
    _fileType = "vcf"

    _process_kwargs = []

    def _validateFilename(self, filePath):
        basename = os.path.basename(filePath[0])
        startswith_genie = basename.startswith("GENIE-{}-".format(self.center))
        endswith_vcf = basename.endswith(".vcf")
        assert startswith_genie and endswith_vcf

    def _get_dataframe(self, filePathList):
        headers = None
        filepath = filePathList[0]
        with open(filepath, "r") as vcffile:
            for row in vcffile:
                if row.startswith("#CHROM"):
                    headers = row.replace("\n", "").replace("\r", "").split("\t")
                    break
        if headers is not None:
            vcfdf = pd.read_csv(
                filepath, sep="\t", comment="#", header=None, names=headers
            )
        else:
            raise ValueError("Your vcf must start with the header #CHROM")
        return vcfdf

    def process_steps(self, df):
        """The processing of vcf files is specific to GENIE, so
        not included in this function"""
        logger.info(
            "Please run with `--process mutation` parameter "
            "if you want to reannotate the mutation files"
        )
        return None

    def _validate(self, vcfdf):
        """
        Validates the content of a vcf file

        Args:
            vcfdf: pandas dataframe containing vcf content

        Returns:
            total_error - error messages
            warning - warning messages
        """
        required_headers = pd.Series(
            ["#CHROM", "POS", "ID", "REF", "ALT", "QUAL", "FILTER", "INFO"]
        )
        total_error = ""
        warning = ""
        if not all(required_headers.isin(vcfdf.columns)):
            total_error += (
                "vcf: Must have these headers: "
                "CHROM, POS, ID, REF, ALT, QUAL, FILTER, INFO.\n"
            )
        else:
            # No duplicated values
            primary_cols = ["#CHROM", "POS", "REF", "ALT"]
            if vcfdf.duplicated(primary_cols).any():
                total_error += "vcf: Must not have duplicate variants.\n"

            if vcfdf[["#CHROM", "POS"]].isnull().values.any():
                total_error += (
                    "vcf: May contain rows that are "
                    "space delimited instead of tab delimited.\n"
                )

        if len(vcfdf.columns) > 8:
            if "FORMAT" not in vcfdf.columns:
                total_error += (
                    "vcf: Must have FORMAT header " "if genotype columns exist.\n"
                )

        # Require that they report variants mapped to
        # either GRCh37 or hg19 without
        # the chr-prefix. variants on chrM are not supported
        have_column = process_functions.checkColExist(vcfdf, "#CHROM")
        if have_column:
            nochr = ["chr" in i for i in vcfdf["#CHROM"] if isinstance(i, str)]
            if sum(nochr) > 0:
                warning += (
                    "vcf: Should not have the chr prefix " "in front of chromosomes.\n"
                )
            if sum(vcfdf["#CHROM"].isin(["chrM"])) > 0:
                total_error += "vcf: Must not have variants on chrM.\n"

        # No white spaces
        white_space = vcfdf.apply(lambda x: contains_whitespace(x), axis=1)
        if sum(white_space) > 0:
            warning += (
                "vcf: Should not have any " "white spaces in any of the columns.\n"
            )

        # I can also recommend a `bcftools query` command that
        # will parse a VCF in a detailed way,
        # and output with warnings or errors if the format is not adhered too
        return total_error, warning



from unittest import mock

import pandas as pd
import pytest
import synapseclient

from genie_registry.seg import seg
from genie_registry.cbs import cbs

syn = mock.create_autospec(synapseclient.Synapse)

segClass = seg(syn, "SAGE")
cbsClass = cbs(syn, "SAGE")


def test_processing():
    expectedSegDf = pd.DataFrame(
        {
            "ID": [
                "GENIE-SAGE-ID1-1",
                "GENIE-SAGE-ID2-1",
                "GENIE-SAGE-ID3-1",
                "GENIE-SAGE-ID4-1",
                "GENIE-SAGE-ID5-1",
            ],
            "CHROM": ["1", "2", "3", "4", "5"],
            "LOCSTART": [1, 2, 3, 4, 3],
            "LOCEND": [1, 2, 3, 4, 2],
            "NUMMARK": [1, 2, 3, 4, 3],
            "SEGMEAN": [1, 2, 3.9, 4, 3],
            "CENTER": ["SAGE", "SAGE", "SAGE", "SAGE", "SAGE"],
        }
    )

    segDf = pd.DataFrame(
        {
            "ID": [
                "GENIE-SAGE-ID1-1",
                "GENIE-SAGE-ID2-1",
                "GENIE-SAGE-ID3-1",
                "GENIE-SAGE-ID4-1",
                "GENIE-SAGE-ID5-1",
            ],
            "CHROM": ["chr1", 2, 3, 4, 5],
            "LOC.START": [1, 2, 3, 4, 3],
            "LOC.END": [1, 2, 3, 4, 2],
            "NUM.MARK": [1, 2, 3, 4, 3],
            "SEG.MEAN": [1, 2, 3.9, 4, 3],
        }
    )

    newSegDf = segClass._process(segDf)
    assert expectedSegDf.equals(newSegDf[expectedSegDf.columns])

    newSegDf = cbsClass._process(segDf)
    assert expectedSegDf.equals(newSegDf[expectedSegDf.columns])


def test_validation_filename():
    with pytest.raises(AssertionError):
        segClass.validateFilename(["foo"])
    assert segClass.validateFilename(["genie_data_cna_hg19_SAGE.seg"]) == "seg"
    assert cbsClass.validateFilename(["genie_data_cna_hg19_SAGE.cbs"]) == "cbs"


def test_validation_perfect():
    segDf = pd.DataFrame(
        {
            "ID": [
                "GENIE-SAGE-ID1",
                "GENIE-SAGE-ID2",
                "GENIE-SAGE-ID3",
                "GENIE-SAGE-ID4",
                "GENIE-SAGE-ID5",
            ],
            "CHROM": [1, 2, 3, 4, 5],
            "LOC.START": [1, 2, 3, 4, 3],
            "LOC.END": [1, 2, 3, 4, 3],
            "NUM.MARK": [1, 2, 3, 4, 3],
            "SEG.MEAN": [1, 2, 3, 4, 3],
        }
    )

    error, warning = segClass._validate(segDf)
    assert error == ""
    assert warning == ""


def test_valdation_invalid():
    segDf = pd.DataFrame(
        {
            "ID": ["ID1", "ID2", "ID3", "ID4", "ID5"],
            "CHROM": [1, 2, float("nan"), 4, 5],
            "LOC.START": [1, 2, 3, 4, float("nan")],
            "LOC.END": [1, 2, 3, float("nan"), 3],
            "NUM.MARK": [1, 2, 3, 4, 3],
        }
    )
    expectedErrors = (
        "Your seg file is missing these headers: SEG.MEAN.\n"
        "Seg: No null or empty values allowed in column(s): "
        "CHROM, LOC.END, LOC.START.\n"
        "Seg: ID must start with GENIE-SAGE\n"
    )
    error, warning = segClass._validate(segDf)
    assert error == expectedErrors
    assert warning == ""

    error, warning = cbsClass._validate(segDf)
    assert error == expectedErrors
    assert warning == ""

    segDf = pd.DataFrame(
        {
            "ID": ["ID1", "ID2", "ID3", "ID4", "ID5"],
            "CHROM": [1, 2, 3, 4, 5],
            "LOC.START": [1, 2, 3, 4.3, 3],
            "LOC.END": [1, 2, 3.4, 4, 3],
            "NUM.MARK": [1, 2, 3, 33.3, 3],
            "SEG.MEAN": [1, 2, "f.d", 4, 3],
        }
    )
    error, warning = segClass._validate(segDf)
    expectedErrors = (
        "Seg: Only integars allowed in these column(s): "
        "LOC.END, LOC.START, NUM.MARK.\n"
        "Seg: Only numerical values allowed in SEG.MEAN.\n"
        "Seg: ID must start with GENIE-SAGE\n"
    )
    assert error == expectedErrors
    assert warning == ""



"""GENIE bed class and functions"""
import os
import logging
import subprocess

import pandas as pd

from genie.example_filetype_format import FileTypeFormat
from genie import process_functions

LOGGER = logging.getLogger(__name__)


# def createGenePositionsTables():
#     '''
#     Create gene position database
#     '''
#     from biomart import BiomartServer
#     import synapseclient
#     import time
#     syn = synapseclient.login()
#     print(time.time())
#     server = BiomartServer("grch37.ensembl.org/biomart")

#     hsapiens_gene_ensembl = server.datasets['hsapiens_gene_ensembl']
#     # hsapiens_gene_ensembl.show_attributes()

#     chromosomes = range(1, 23)
#     chromosomes.extend(["X", "Y"])
#     chromosomes = [str(i) for i in chromosomes]
#     response = hsapiens_gene_ensembl.search({
#         'attributes': [
#             'hgnc_symbol',
#             "chromosome_name",
#             "start_position",
#             "end_position"]
#     })
#     genes = pd.DataFrame()
#     for line in response.iter_lines():
#         line = line.decode('utf-8')
#         row = line.split("\t")
#         if row[0] != '' and row[1] in chromosomes:
#             tempRow = pd.DataFrame([row])
#             genes = genes.append(tempRow)
#     genes.columns = [
#         'hgnc_symbol', "chromosome_name",
#         "start_position", "end_position"]
#     # genes.to_csv("genepositions.csv",index=False)
#     print(time.time())
#     databaseSynId = "syn11806563"
#     databaseEnt = syn.get(databaseSynId)
#     database = syn.tableQuery("SELECT * FROM %s" % databaseSynId)
#     database = database.asDataFrame()
#     process_functions.updateDatabase(
#         syn, database, genes, databaseSynId,
#         databaseEnt.primaryKey, toDelete=True)
#     return(genes)


def create_gtf(dirname):
    """
    Create exon.gtf and gene.gtf from GRCh37 gtf

    Args:
        dirname: Directory where these files should live

    Returns:
        exon_gtf_path: exon GTF
        gene_gtf_path: gene GTF
    """
    exon_gtf_path = os.path.join(dirname, "exon.gtf")
    gene_gtf_path = os.path.join(dirname, "gene.gtf")

    if not os.path.exists(exon_gtf_path) or not os.path.exists(gene_gtf_path):
        download_cmd = [
            "wget",
            "http://ftp.ensembl.org/pub/release-75/gtf/homo_sapiens/Homo_sapiens.GRCh37.75.gtf.gz",
            "-P",
            dirname,
        ]
        subprocess.check_call(download_cmd)
        gtfgz_path = os.path.join(dirname, "Homo_sapiens.GRCh37.75.gtf.gz")
        gunzip_cmd = ["gunzip", "-f", gtfgz_path]
        subprocess.check_call(gunzip_cmd)
        gtf_path = os.path.join(dirname, "Homo_sapiens.GRCh37.75.gtf")

        exon_awk_cmd = ["awk", '$3 == "exon" {print}', gtf_path]
        exon_gtf = subprocess.check_output(exon_awk_cmd, universal_newlines=True)
        with open(exon_gtf_path, "w") as gtf_file:
            gtf_file.write(exon_gtf)
        gene_awk_cmd = ["awk", '$3 == "gene" {print}', gtf_path]
        gene_gtf = subprocess.check_output(gene_awk_cmd, universal_newlines=True)
        with open(gene_gtf_path, "w") as gtf_file:
            gtf_file.write(gene_gtf)
    return (exon_gtf_path, gene_gtf_path)


def _add_feature_type_tobeddf(filepath, featuretype):
    """
    Add Feature_Type to dataframe

    Args:
        filepath: path to bed
        featuretype: exon, intron, or intergenic

    Returns:
        df: empty dataframe or dataframe with appended feature type
    """
    bed_columns = [
        "Chromosome",
        "Start_Position",
        "End_Position",
        "Hugo_Symbol",
        "includeInPanel",
        "clinicalReported",
        "ID",
        "SEQ_ASSAY_ID",
        "Feature_Type",
    ]
    # No need to add anything if the dataframe is empty
    if os.stat(filepath).st_size != 0:
        beddf = pd.read_csv(filepath, sep="\t", header=None)
        beddf["Feature_Type"] = featuretype
        beddf.columns = bed_columns
    else:
        beddf = pd.DataFrame(columns=bed_columns)
    return beddf


def add_feature_type(temp_bed_path, exon_gtf_path, gene_gtf_path):
    """
    Add Feature_Type to bed file (exon, intron, intergenic)

    Args:
        temp_bed_path: BED file without feature type
        exon_gtf_path: exon gtf
        gene_gtf_path: gene gtf

    Returns:
        genie_combined_path: Path to final bed file
    """
    genie_exon_path = os.path.join(process_functions.SCRIPT_DIR, "genie_exons.bed")
    genie_intron_path = os.path.join(process_functions.SCRIPT_DIR, "genie_introns.bed")
    genie_intergenic_path = os.path.join(
        process_functions.SCRIPT_DIR, "genie_intergenic.bed"
    )
    intron_intergenic_path = os.path.join(
        process_functions.SCRIPT_DIR, "intron_intergenic.bed"
    )
    gene_path = os.path.join(process_functions.SCRIPT_DIR, "gene.bed")
    # GET EXON REGIONS
    # Get intersection between true exon regions and submitted bed regions
    command = [
        "bedtools",
        "intersect",
        "-a",
        temp_bed_path,
        "-b",
        exon_gtf_path,
        "-wa",
        "|",
        "sort",
        "|",
        "uniq",
        ">",
        genie_exon_path,
    ]
    subprocess.check_call(" ".join(command), shell=True)
    # get intergenic/intron regions
    # Get opposite of intersection between true exon regions and submitted
    # bed regions
    command = [
        "bedtools",
        "intersect",
        "-a",
        temp_bed_path,
        "-b",
        exon_gtf_path,
        "-wa",
        "-v",
        "|",
        "sort",
        "|",
        "uniq",
        ">",
        intron_intergenic_path,
    ]
    subprocess.check_call(" ".join(command), shell=True)
    # get gene regions
    # Get intersection between true gene regions and submitted bed regions
    command = [
        "bedtools",
        "intersect",
        "-a",
        temp_bed_path,
        "-b",
        gene_gtf_path,
        "-wa",
        "|",
        "sort",
        "|",
        "uniq",
        ">",
        gene_path,
    ]
    subprocess.check_call(" ".join(command), shell=True)
    # GET INTRON REGSIONS
    # Difference between the gene regions and exon regions will give
    # intron regions
    command = [
        "diff",
        gene_path,
        genie_exon_path,
        "|",
        "grep",
        "'<'",
        "|",
        "sed",
        "'s/< //'",
        ">",
        genie_intron_path,
    ]
    subprocess.check_call(" ".join(command), shell=True)
    # GET INTERGENIC REGIONS
    # Difference between the intron/intergenic and intron regions will
    # give intergenic regions
    command = [
        "diff",
        intron_intergenic_path,
        genie_intron_path,
        "|",
        "grep",
        "'<'",
        "|",
        "sed",
        "'s/< //'",
        ">",
        genie_intergenic_path,
    ]
    subprocess.check_call(" ".join(command), shell=True)

    genie_exondf = _add_feature_type_tobeddf(genie_exon_path, "exon")
    genie_introndf = _add_feature_type_tobeddf(genie_intron_path, "intron")
    genie_intergenicdf = _add_feature_type_tobeddf(genie_intergenic_path, "intergenic")
    genie_combineddf = pd.concat([genie_exondf, genie_introndf, genie_intergenicdf])
    return genie_combineddf


def _check_region_overlap(row, gene_positiondf):
    """
    Check if the submitted bed symbol + region overlaps with
    the actual gene's positions

    Args:
        row: row in bed file (genomic region)
        gene_positiondf: Reference gene position dataframe

    Return:
        True if the region does overlap.
    """
    matching_symbol_ind = gene_positiondf["hgnc_symbol"] == row["Hugo_Symbol"]
    match_with_genedb = gene_positiondf[matching_symbol_ind]

    if not match_with_genedb.empty:
        # We are assuming that there is only one matching gene, but
        # there are actually duplicated gene symbols
        start_position = match_with_genedb["start_position"].values[0]
        end_position = match_with_genedb["end_position"].values[0]
        # Submitted bed start position in a gene
        start_in_gene = start_position <= row["Start_Position"] <= end_position
        # Submitted bed end position in a gene
        end_in_gene = start_position <= row["End_Position"] <= end_position
        # Submitted bed region surrounds a gene
        gene_in_region = (
            end_position <= row["End_Position"]
            and start_position >= row["Start_Position"]
        )
        return start_in_gene or end_in_gene or gene_in_region
    return False


def _get_max_overlap_index(overlap, bed_length, boundary):
    """
    Calculate the ratio of overlap between the submitted bed
    region and gene position dataframe and return the index of
    the max overlapping region

    Args:
        overlap: Possible overlapping region
        bed_length: Length of submitted region
        boundary: specified ratio overlap

    Returns:
        Index of regions with maximum overlap or None
    """
    ratio_overlap = overlap / bed_length
    ratio_overlap = ratio_overlap[ratio_overlap > boundary]
    ratio_overlap = ratio_overlap[ratio_overlap <= 1]
    if not ratio_overlap.empty:
        return ratio_overlap.idxmax()
    return None


def _map_position_within_boundary(row, positiondf, boundary=0.9):
    """
    Map positions and checks if posision is contained within the
    specified percentage boundary

    Args:
        row: Row in bed file (genomic region)
        positiondf: Reference bed position dataframe
        boundary: Percent boundary defined

    Return:
        pd.Series: mapped position
    """
    # First filter just based on whether or not region is within the gene
    # position dataframe
    match_chr = positiondf["chromosome_name"] == str(row["Chromosome"])
    chrom_rows = positiondf[match_chr]
    match_start = chrom_rows["start_position"] <= row["Start_Position"]
    start_rows = chrom_rows[match_start]
    end_rows = start_rows[start_rows["end_position"] >= row["End_Position"]]

    bed_length = row["End_Position"] - row["Start_Position"]
    # as long as the strand is within the boundary % defined
    # Start goes over start boundary, but end is contained in position
    if end_rows.empty:
        if sum(chrom_rows["end_position"] >= row["End_Position"]) > 0:
            overlap = row["End_Position"] - chrom_rows["start_position"]
            # difference =  difference * -1.0
            max_overlap = _get_max_overlap_index(overlap, bed_length, boundary)
            if max_overlap is not None:
                end_rows = pd.concat([end_rows, chrom_rows.loc[[max_overlap]]])
        # End goes over end boundary, but start is contained in position
        if sum(chrom_rows["start_position"] <= row["Start_Position"]) > 0:
            overlap = chrom_rows["end_position"] - row["Start_Position"]
            max_overlap = _get_max_overlap_index(overlap, bed_length, boundary)
            if max_overlap is not None:
                end_rows = pd.concat([end_rows, chrom_rows.loc[[max_overlap]]])
        # Start and end go over position boundary
        check = chrom_rows[chrom_rows["start_position"] >= row["Start_Position"]]
        check = check[check["end_position"] <= row["End_Position"]]
        if not check.empty:
            overlap = chrom_rows["end_position"] - chrom_rows["start_position"]
            max_overlap = _get_max_overlap_index(overlap, bed_length, boundary)
            if max_overlap is not None:
                end_rows = pd.concat([end_rows, chrom_rows.loc[[max_overlap]]])
    return end_rows


def remap_symbols(row, gene_positiondf):
    """
    Remap hugo symbols if there is no overlap between submitted bed region and
    gene positions.

    Args:
        row: start and end position
        gene_positiondf: Actual gene position dataframe

    Return:
        bool or Series: if the gene passed in need to be remapped or
                        the remapped gene
    """
    region_overlap = _check_region_overlap(row, gene_positiondf)
    if not region_overlap:
        overlap_positions = _map_position_within_boundary(row, gene_positiondf)
        if overlap_positions.empty:
            LOGGER.warning(
                "{} cannot be remapped. "
                "These rows will have an empty gene symbol".format(row["Hugo_Symbol"])
            )
            row["Hugo_Symbol"] = float("nan")
        elif len(overlap_positions) > 1:
            symbol_list = overlap_positions["hgnc_symbol"].tolist()
            if row["Hugo_Symbol"] not in symbol_list:
                # if "MLL4", then the HUGO symbol should be KMT2D and KMT2B
                LOGGER.warning(
                    "{} can be mapped to different symbols: {}. "
                    "Please correct or it will be removed.".format(
                        row["Hugo_Symbol"], ", ".join(symbol_list)
                    )
                )
                row["Hugo_Symbol"] = float("nan")
        else:
            symbol = overlap_positions["hgnc_symbol"].values[0]
            if row["Hugo_Symbol"] != symbol:
                LOGGER.info(
                    "{} will be remapped to {}".format(row["Hugo_Symbol"], symbol)
                )
                row["Hugo_Symbol"] = symbol
    return row


class bed(FileTypeFormat):
    """GENIE bed format"""

    _fileType = "bed"

    _process_kwargs = ["newPath", "parentId", "databaseSynId", "seq_assay_id"]

    def _get_dataframe(self, filepathlist):
        """
        Bed files don't have a header

        Args:
            filePathList: List of files
        """
        filepath = filepathlist[0]
        try:
            beddf = pd.read_csv(filepath, sep="\t", header=None)
        except Exception:
            raise ValueError(
                "Can't read in your bed file. "
                "Please make sure the BED file is not binary and "
                "does not contain a comment/header line"
            )
        first_value = str(beddf[0][0])
        if (
            not first_value.isdigit()
            and not first_value.startswith("chr")
            and first_value not in ["X", "Y"]
        ):
            raise ValueError(
                "Please make sure your bed file does not "
                "contain a comment/header line"
            )
        return beddf

    def _validateFilename(self, filepath):
        """
        Validates filename
        CENTER-11.bed

        Args:
            filePath: Path to bedfile
        """

        assert os.path.basename(filepath[0]).startswith(
            "%s-" % self.center
        ) and os.path.basename(filepath[0]).endswith(".bed")

    def create_gene_panel(self, beddf, seq_assay_id, gene_panel_path, parentid):
        """
        Create bed file and gene panel files from the bed file

        Args:
            beddf: bed dataframe
            seq_assay_id: GENIE SEQ_ASSAY_ID
            gene_panel_path: Gene panel folder path
            parentid: Synapse id of gene panel folder

        Returns:
            pd.DataFrame: configured bed dataframe
        """
        LOGGER.info("CREATING GENE PANEL")
        if not beddf.empty:
            exonsdf = beddf[beddf["Feature_Type"] == "exon"]
            # Only include genes that should be included in the panels
            include_exonsdf = exonsdf[exonsdf["includeInPanel"]]
            # Write gene panel
            null_genes = include_exonsdf["Hugo_Symbol"].isnull()
            unique_genes = set(include_exonsdf["Hugo_Symbol"][~null_genes])
            gene_panel_text = (
                "stable_id: {seq_assay_id}\n"
                "description: {seq_assay_id}, "
                "Number of Genes - {num_genes}\n"
                "gene_list:\t{genelist}".format(
                    seq_assay_id=seq_assay_id,
                    num_genes=len(unique_genes),
                    genelist="\t".join(unique_genes),
                )
            )
            gene_panel_name = "data_gene_panel_" + seq_assay_id + ".txt"

            with open(os.path.join(gene_panel_path, gene_panel_name), "w+") as f:
                f.write(gene_panel_text)
            process_functions.storeFile(
                self.syn,
                os.path.join(gene_panel_path, gene_panel_name),
                parentId=parentid,
                center=self.center,
                fileFormat="bed",
                dataSubType="metadata",
                cBioFileFormat="genePanel",
            )

    def _process(self, beddf, seq_assay_id, newpath, parentid, create_panel=True):
        """
        Process bed file, add feature type

        Args:
            gene: bed dataframe
            seq_assay_id: GENIE SEQ_ASSAY_ID
            newpath: new GENIE path
            parentid: Synapse id to store the gene panel
            create_panel: Create gene panel

        Returns:
            pd.DataFrame: Conigured bed dataframe
        """
        seq_assay_id = seq_assay_id.upper()
        seq_assay_id = seq_assay_id.replace("_", "-")

        # Add in 6th column which is the clinicalReported
        if len(beddf.columns) > 5:
            if all(beddf[5].apply(lambda x: x in [True, False])):
                beddf[5] = beddf[5].astype(bool)
            else:
                beddf[5] = float("nan")
        else:
            beddf[5] = float("nan")
        beddf = beddf[[0, 1, 2, 3, 4, 5]]
        gene_panel_path = os.path.dirname(newpath)
        # Must be .astype(bool) because `1, 0 in [True, False]`
        beddf[4] = beddf[4].astype(bool)

        exon_gtf_path, gene_gtf_path = create_gtf(process_functions.SCRIPT_DIR)
        LOGGER.info("REMAPPING {}".format(seq_assay_id))
        # bedname = seq_assay_id + ".bed"
        beddf.columns = [
            "Chromosome",
            "Start_Position",
            "End_Position",
            "Hugo_Symbol",
            "includeInPanel",
            "clinicalReported",
        ]
        # Validate gene symbols
        # Gene symbols can be split by ; and _ and : and .
        beddf["Hugo_Symbol"] = [
            i.split(";")[0].split("_")[0].split(":")[0].split(".")[0]
            for i in beddf["Hugo_Symbol"]
        ]
        # Replace all chr with blank
        beddf["Chromosome"] = [str(i).replace("chr", "") for i in beddf["Chromosome"]]
        # Change all start and end to int
        beddf["Start_Position"] = beddf["Start_Position"].apply(int)
        beddf["End_Position"] = beddf["End_Position"].apply(int)

        gene_position_table = self.syn.tableQuery("SELECT * FROM syn11806563")
        gene_positiondf = gene_position_table.asDataFrame()
        beddf["ID"] = beddf["Hugo_Symbol"]
        # The apply function of a DataFrame is called twice on the first
        # row (known pandas behavior)
        beddf = beddf.apply(lambda x: remap_symbols(x, gene_positiondf), axis=1)
        beddf["SEQ_ASSAY_ID"] = seq_assay_id
        temp_bed_path = os.path.join(process_functions.SCRIPT_DIR, "temp.bed")
        beddf.to_csv(temp_bed_path, sep="\t", index=False, header=None)
        final_bed = add_feature_type(temp_bed_path, exon_gtf_path, gene_gtf_path)
        final_bed["CENTER"] = self.center
        final_bed["Chromosome"] = final_bed["Chromosome"].astype(str)
        if create_panel:
            self.create_gene_panel(final_bed, seq_assay_id, gene_panel_path, parentid)
        return final_bed

    def preprocess(self, newpath):
        """
        Standardize and grab seq assay id from the bed file path

        Args:
            newpath: bed file path

        Returns:
            dict: GENIE seq assay id
        """
        seq_assay_id = os.path.basename(newpath).replace(".bed", "")
        seq_assay_id = seq_assay_id.upper().replace("_", "-")
        return {"seq_assay_id": seq_assay_id}

    def process_steps(self, beddf, newPath, parentId, databaseSynId, seq_assay_id):
        """
        Process bed file, update bed database, write bed file to path

        Args:
            beddf: Bed dataframe
            newPath: Path to new bed file
            parentId: Synapse id to store gene panel file
            databaseSynId: Synapse id of bed database
            seq_assay_id: GENIE seq assay id

        Returns:
            string: Path to new bed file
        """
        final_beddf = self._process(beddf, seq_assay_id, newPath, parentId)
        process_functions.updateData(
            self.syn,
            databaseSynId,
            final_beddf,
            seq_assay_id,
            filterByColumn="SEQ_ASSAY_ID",
            toDelete=True,
        )
        final_beddf.to_csv(newPath, sep="\t", index=False)
        return newPath

    def _validate(self, beddf):
        """
        Validate bed file

        Args:
            bed: Bed dataframe

        Returns:
            total_error: all the errors
            warning: all the warnings
        """
        total_error = ""
        warning = ""
        newcols = [
            "Chromosome",
            "Start_Position",
            "End_Position",
            "Hugo_Symbol",
            "includeInPanel",
        ]
        if len(beddf.columns) < len(newcols):
            total_error += (
                "BED file: Must at least have five columns in this "
                "order: {}. Make sure there are "
                "no headers.\n".format(", ".join(newcols))
            )
        else:
            newcols.extend(range(0, len(beddf.columns) - len(newcols)))
            beddf.columns = newcols
            to_validate_symbol = True
            if not all(beddf["Start_Position"].apply(lambda x: isinstance(x, int))):
                total_error += (
                    "BED file: "
                    "The Start_Position column must only be "
                    "integers. Make sure there are no headers.\n"
                )
                to_validate_symbol = False
            if not all(beddf["End_Position"].apply(lambda x: isinstance(x, int))):
                total_error += (
                    "BED file: "
                    "The End_Position column must only be "
                    "integers. Make sure there are no headers.\n"
                )
                to_validate_symbol = False

            LOGGER.info("VALIDATING GENE SYMBOLS")
            if any(beddf["Hugo_Symbol"].isnull()):
                total_error += "BED file: You cannot submit any null symbols.\n"
            beddf = beddf[~beddf["Hugo_Symbol"].isnull()]
            beddf["Hugo_Symbol"] = [
                str(hugo).split(";")[0].split("_")[0].split(":")[0]
                for hugo in beddf["Hugo_Symbol"]
            ]
            if (
                sum(beddf["Hugo_Symbol"] == "+") != 0
                or sum(beddf["Hugo_Symbol"] == "-") != 0
            ):
                total_error += (
                    "BED file: Fourth column must be the "
                    "Hugo_Symbol column, not the strand column\n"
                )

            warn, error = process_functions.check_col_and_values(
                beddf,
                "includeInPanel",
                [True, False],
                filename="BED file",
                required=True,
            )
            warning += warn
            total_error += error

            if to_validate_symbol:
                gene_position_table = self.syn.tableQuery("SELECT * FROM syn11806563")
                gene_positiondf = gene_position_table.asDataFrame()
                # The apply function of a DataFrame is called twice on the first row (known
                # pandas behavior)
                beddf = beddf.apply(lambda x: remap_symbols(x, gene_positiondf), axis=1)

                if any(beddf["Hugo_Symbol"].isnull()):
                    warning += (
                        "BED file: "
                        "Any gene names that can't be "
                        "remapped will be null.\n"
                    )
                if all(beddf["Hugo_Symbol"].isnull()):
                    total_error += (
                        "BED file: "
                        "You have no correct gene symbols. "
                        "Make sure your gene symbol column (4th "
                        "column) is formatted like so: SYMBOL"
                        "(;optionaltext).  Optional text can be "
                        "semi-colon separated.\n"
                    )
        return (total_error, warning)



"""Tests validate.py"""
from unittest import mock
from unittest.mock import Mock, patch

import pandas as pd
import pytest
import synapseclient
from synapseclient.core.exceptions import SynapseHTTPError

from genie import validate, process_functions, example_filetype_format

CENTER = "SAGE"
syn = mock.create_autospec(synapseclient.Synapse)

CNA_ENT = synapseclient.File(
    name="data_CNA_SAGE.txt", path="data_CNA_SAGE.txt", parentId="syn12345"
)
CLIN_ENT = synapseclient.File(
    name="data_clinical_supp_SAGE.txt",
    path="data_clinical_supp_SAGE.txt",
    parentId="syn12345",
)
SAMPLE_ENT = synapseclient.File(
    name="data_clinical_supp_sample_SAGE.txt",
    path="data_clinical_supp_sample_SAGE.txt",
    parentId="syn12345",
)
PATIENT_ENT = synapseclient.File(
    name="data_clinical_supp_patient_SAGE.txt",
    path="data_clinical_supp_patient_SAGE.txt",
    parentId="syn12345",
)
WRONG_NAME_ENT = synapseclient.File(
    name="wrong.txt", path="data_clinical_supp_SAGE.txt", parentId="syn12345"
)


class FileFormat(example_filetype_format.FileTypeFormat):
    _fileType = "clinical"


def test_perfect_determine_filetype():
    """
    Tests determining of file type through filenames
    Parameters are passed in from filename_fileformat_map
    """
    filetype = "clincial"
    ent_list = [SAMPLE_ENT]
    with patch.object(FileFormat, "validateFilename", return_value=filetype):
        validator = validate.GenieValidationHelper(
            syn, None, CENTER, ent_list, format_registry={filetype: FileFormat}
        )
        assert validator.determine_filetype() == filetype


def test_wrongfilename_noerror_determine_filetype():
    """
    Tests None is passed back when wrong filename is passed
    when raise_error flag is False
    """
    ent_list = [WRONG_NAME_ENT]
    with patch.object(FileFormat, "validateFilename", side_effect=AssertionError):
        validator = validate.GenieValidationHelper(
            syn,
            project_id=None,
            center=CENTER,
            entitylist=ent_list,
            format_registry={"wrong": FileFormat},
        )
        assert validator.file_type is None


def test_valid_collect_errors_and_warnings():
    """
    Tests if no error and warning strings are passed that
    returned valid and message is correct
    """
    results = example_filetype_format.ValidationResults(errors="", warnings="")
    message = results.collect_errors_and_warnings()
    assert message == "YOUR FILE IS VALIDATED!\n"


def test_invalid_collect_errors_and_warnings():
    """
    Tests if error and warnings strings are passed that
    returned valid and message is correct
    """
    results = example_filetype_format.ValidationResults(
        errors="error\nnow", warnings="warning\nnow"
    )
    message = results.collect_errors_and_warnings()
    assert message == (
        "----------------ERRORS----------------\n"
        "error\nnow"
        "-------------WARNINGS-------------\n"
        "warning\nnow"
    )


def test_warning_collect_errors_and_warnings():
    """
    Tests if no error but warnings strings are passed that
    returned valid and message is correct
    """
    results = example_filetype_format.ValidationResults(
        errors="", warnings="warning\nnow"
    )
    message = results.collect_errors_and_warnings()
    assert message == (
        "YOUR FILE IS VALIDATED!\n"
        "-------------WARNINGS-------------\n"
        "warning\nnow"
    )


def test_valid_validate_single_file(genie_config):
    """
    Tests that all the functions are run in validate single
    file workflow and all the right things are returned
    """
    entitylist = [CLIN_ENT]
    results = example_filetype_format.ValidationResults(errors="", warnings="")
    expected_message = "valid message here!"
    expected_filetype = "clinical"
    project_ent = Mock(id="syn1234")
    with patch.object(syn, "get", return_value=project_ent), patch.object(
        validate.GenieValidationHelper,
        "determine_filetype",
        return_value=expected_filetype,
    ) as mock_determine_ftype, patch.object(
        FileFormat, "validate", return_value=results
    ) as mock_genie_class, patch.object(
        results, "collect_errors_and_warnings", return_value=expected_message
    ) as mock_determine:
        validator = validate.GenieValidationHelper(
            syn,
            project_id="syn1234",
            center=CENTER,
            entitylist=entitylist,
            format_registry={"clinical": FileFormat},
            genie_config=genie_config,
        )
        valid_cls, message = validator.validate_single_file(nosymbol_check=False)

        assert valid_cls == results
        assert message == expected_message
        assert validator.file_type == expected_filetype

        mock_determine_ftype.assert_called_once_with()

        mock_genie_class.assert_called_once_with(
            filePathList=[CLIN_ENT.path], nosymbol_check=False, project_id="syn1234"
        )

        mock_determine.assert_called_once_with()


def test_filetype_validate_single_file():
    """
    Tests that if filetype is passed in that an error is thrown
    if it is an incorrect filetype
    """
    entitylist = [WRONG_NAME_ENT]
    expected_error = (
        "----------------ERRORS----------------\n"
        "Your filename is incorrect! Please change your "
        "filename before you run the validator or specify "
        "--filetype if you are running the validator locally"
    )

    with patch.object(FileFormat, "validateFilename", side_effect=AssertionError):
        validator = validate.GenieValidationHelper(
            syn, None, CENTER, entitylist, format_registry={"wrong": FileFormat}
        )

        valid_cls, message = validator.validate_single_file()
        assert message == expected_error
        assert not valid_cls.is_valid()


def test_wrongfiletype_validate_single_file():
    """
    Tests that if there is no filetype for the filename passed
    in, an error is thrown
    """
    entitylist = [WRONG_NAME_ENT]
    expected_error = (
        "----------------ERRORS----------------\n"
        "Your filename is incorrect! Please change your "
        "filename before you run the validator or specify "
        "--filetype if you are running the validator locally"
    )

    with patch.object(
        validate.GenieValidationHelper, "determine_filetype", return_value=None
    ) as mock_determine_filetype:
        validator = validate.GenieValidationHelper(
            syn=syn,
            project_id=None,
            center=CENTER,
            entitylist=entitylist,
            format_registry={"wrong": Mock()},
        )
        valid, message = validator.validate_single_file()

        assert message == expected_error
        assert not valid.is_valid()
        mock_determine_filetype.assert_called_once_with()


def test_nopermission__check_parentid_permission_container():
    """Throws error if no permissions to access"""
    parentid = "syn123"
    with patch.object(syn, "get", side_effect=SynapseHTTPError), pytest.raises(
        ValueError,
        match="Provided Synapse id must be your input folder "
        "Synapse id or a Synapse Id of a folder inside "
        "your input directory",
    ):
        validate._check_parentid_permission_container(syn, parentid)


def test_notcontainer__check_parentid_permission_container():
    """Throws error if input if synid of file"""
    parentid = "syn123"
    file_ent = synapseclient.File("foo", parentId=parentid)
    with patch.object(syn, "get", return_value=file_ent), pytest.raises(
        ValueError,
        match="Provided Synapse id must be your input folder "
        "Synapse id or a Synapse Id of a folder inside "
        "your input directory",
    ):
        validate._check_parentid_permission_container(syn, parentid)


def test_valid__check_parentid_permission_container():
    """
    Test that parentid specified is a container and have permissions to access
    """
    parentid = "syn123"
    folder_ent = synapseclient.Folder("foo", parentId=parentid)
    with patch.object(syn, "get", return_value=folder_ent):
        validate._check_parentid_permission_container(syn, parentid)


def test_valid__check_center_input():
    center = "FOO"
    center_list = ["FOO", "WOW"]
    validate._check_center_input(center, center_list)


def test_invalid__check_center_input():
    center = "BARFOO"
    center_list = ["FOO", "WOW"]
    with pytest.raises(
        ValueError,
        match="Must specify one of these " "centers: {}".format(", ".join(center_list)),
    ):
        validate._check_center_input(center, center_list)


ONCOTREE_ENT = "syn222"


class argparser:
    oncotree_link = "link"
    parentid = None
    filetype = None
    project_id = None
    center = "try"
    filepath = "path.csv"
    nosymbol_check = False
    format_registry_packages = ["genie"]
    project_id = "syn1234"

    def asDataFrame(self):
        database_dict = {
            "Database": ["centerMapping", "oncotreeLink"],
            "Id": ["syn123", ONCOTREE_ENT],
            "center": ["try", "foo"],
        }
        databasetosynid_mappingdf = pd.DataFrame(database_dict)
        return databasetosynid_mappingdf


def test_valid__upload_to_synapse():
    """
    Test upload of file to synapse under right conditions
    """
    ent = synapseclient.File(id="syn123", parentId="syn222")
    with patch.object(syn, "store", return_value=ent) as patch_synstore:
        validate._upload_to_synapse(syn, ["foo"], True, parentid="syn123")
        patch_synstore.assert_called_once_with(
            synapseclient.File("foo", parent="syn123")
        )


def test_perform_validate(genie_config):
    """Make sure all functions are called"""
    arg = argparser()
    valid = True
    with patch.object(
        validate, "_check_parentid_permission_container", return_value=None
    ) as patch_check_parentid, patch.object(
        process_functions, "get_genie_config", return_value=genie_config
    ) as patch_get_config, patch.object(
        validate, "_check_center_input"
    ) as patch_check_center, patch.object(
        process_functions, "_get_oncotreelink"
    ) as patch_get_onco, patch.object(
        validate.GenieValidationHelper,
        "validate_single_file",
        return_value=(valid, "foo"),
    ) as patch_validate, patch.object(
        validate, "_upload_to_synapse"
    ) as patch_syn_upload:
        validate._perform_validate(syn, arg)
        patch_check_parentid.assert_called_once_with(syn=syn, parentid=arg.parentid)
        patch_get_config.assert_called_once_with(syn=syn, project_id=arg.project_id)
        patch_check_center.assert_called_once_with(arg.center, ["SAGE", "TEST"])
        patch_get_onco.assert_called_once()
        patch_validate.assert_called_once_with(
            nosymbol_check=arg.nosymbol_check, project_id=arg.project_id
        )
        patch_syn_upload.assert_called_once_with(
            syn, arg.filepath, valid, parentid=arg.parentid
        )



from .bed import bed
import os
import logging

logger = logging.getLogger(__name__)


class bedSP(bed):

    _fileType = "bedSP"

    def _validateFilename(self, filePath):
        assert os.path.basename(filePath[0]).startswith(
            "nonGENIE_%s-" % self.center
        ) and os.path.basename(filePath[0]).endswith(".bed")



import logging
import os

import pandas as pd
import synapseclient

from genie.example_filetype_format import FileTypeFormat
from genie import process_functions

logger = logging.getLogger(__name__)


def validateSymbol(gene, bedDf, returnMappedDf=True):
    """
    Validate gene symbol

    Args:
        gene: Gene name
        bedDf: Bed pandas dataframe
        returnMappedDf: Return a mapped gene. Defaults to True

    Returns:
        gene name or boolean for whether a gene is valid
    """
    valid = False
    if sum(bedDf["Hugo_Symbol"] == gene) > 0:
        valid = True
    elif sum(bedDf["ID"] == gene) > 0:
        mismatch = bedDf[bedDf["ID"] == gene]
        mismatch.drop_duplicates(inplace=True)
        logger.info(
            "{} will be remapped to {}".format(gene, mismatch["Hugo_Symbol"].values[0])
        )
        gene = mismatch["Hugo_Symbol"].values[0]
    else:
        logger.warning(
            "{} cannot be remapped and will not be released. The symbol "
            "must exist in your seq assay ids (bed files) and must be "
            "mappable to a gene.".format(gene)
        )
        gene = float("nan")
    if returnMappedDf:
        return gene
    else:
        return valid


def makeCNARow(row, symbols):
    """
    Make CNA Row (Deprecated function)

    CNA values are no longer stored in the database

    Args:
        row: one row in the CNA file
        symbols:  list of Gene symbols
    """
    totalrow = "{symbols}\n{values}".format(
        symbols=",".join(symbols), values=",".join(row.astype(str))
    )
    totalrow = totalrow.replace(".0", "")
    return totalrow


def mergeCNAvalues(x):
    """Merge CNA values, make sure if there are two rows that are the
    same gene, the values are merged"""
    # Change into its own series, because sometimes doing an apply
    # will cause there to be a missing index value which will
    # cause dropna() to fail.
    values = pd.Series(x.values)
    values.dropna(inplace=True)
    uniqueValues = set(values.unique())
    if len(uniqueValues) == 1:
        returnVal = x.tolist()[0]
    elif len(uniqueValues) <= 2:
        uniqueValues.discard(0)
        if len(uniqueValues) == 1:
            returnVal = list(uniqueValues)[0]
        else:
            returnVal = float("nan")
    else:
        returnVal = float("nan")
    return returnVal


def checkIfOneZero(x):
    assert len(set(x.tolist())) == 1, "Can only be one unique value"


class cna(FileTypeFormat):

    _fileType = "cna"

    _process_kwargs = ["newPath"]

    _validation_kwargs = ["nosymbol_check"]

    # VALIDATE FILENAME
    def _validateFilename(self, filePath):
        assert os.path.basename(filePath[0]) == "data_CNA_{}.txt".format(self.center)

    def _process(self, cnaDf):
        cnaDf.rename(columns={cnaDf.columns[0]: cnaDf.columns[0].upper()}, inplace=True)
        cnaDf.rename(columns={"HUGO_SYMBOL": "Hugo_Symbol"}, inplace=True)

        index = [
            i for i, col in enumerate(cnaDf.columns) if col.upper() == "ENTREZ_GENE_ID"
        ]
        if len(index) > 0:
            del cnaDf[cnaDf.columns[index][0]]

        bedSynId = self.genie_config["bed"]
        bed = self.syn.tableQuery(
            f"select Hugo_Symbol, ID from {bedSynId} where CENTER = '{self.center}'"
        )
        bedDf = bed.asDataFrame()
        cnaDf["Hugo_Symbol"] = cnaDf["Hugo_Symbol"].apply(
            lambda x: validateSymbol(x, bedDf)
        )
        order = cnaDf.columns
        cnaDf = cnaDf[~cnaDf["Hugo_Symbol"].isnull()]
        # cnaDf = cnaDf.applymap(str)
        duplicatedGenes = pd.DataFrame()
        duplicated_symbols = cnaDf["Hugo_Symbol"][
            cnaDf["Hugo_Symbol"].duplicated()
        ].unique()
        for i in duplicated_symbols:
            dups = cnaDf[cnaDf["Hugo_Symbol"] == i]
            newVal = dups[dups.columns[dups.columns != "Hugo_Symbol"]].apply(
                mergeCNAvalues
            )
            temp = pd.DataFrame(newVal).transpose()
            temp["Hugo_Symbol"] = i
            duplicatedGenes = pd.concat([duplicatedGenes, temp], sort=False)
        cnaDf.drop_duplicates("Hugo_Symbol", keep=False, inplace=True)
        cnaDf = pd.concat([cnaDf, duplicatedGenes], sort=False)
        cnaDf = cnaDf[order]
        return cnaDf

    def process_steps(self, cnaDf, newPath):
        newCNA = self._process(cnaDf)

        centerMafSynId = self.genie_config["centerMaf"]
        if not newCNA.empty:
            cnaText = process_functions.removePandasDfFloat(newCNA)
            # Replace blank with NA's
            cnaText = (
                cnaText.replace("\t\t", "\tNA\t")
                .replace("\t\t", "\tNA\t")
                .replace("\t\n", "\tNA\n")
            )
            with open(newPath, "w") as cnaFile:
                cnaFile.write(cnaText)
            self.syn.store(synapseclient.File(newPath, parent=centerMafSynId))
        return newPath

    def _validate(self, cnvDF, nosymbol_check):
        total_error = ""
        warning = ""
        cnvDF.columns = [col.upper() for col in cnvDF.columns]

        if cnvDF.columns[0] != "HUGO_SYMBOL":
            total_error += "Your cnv file's first column must be Hugo_Symbol\n"
        haveColumn = process_functions.checkColExist(cnvDF, "HUGO_SYMBOL")
        if haveColumn:
            keepSymbols = cnvDF["HUGO_SYMBOL"]
            cnvDF.drop("HUGO_SYMBOL", axis=1, inplace=True)

        # if sum(cnvDF.apply(lambda x: sum(x.isnull()))) > 0:
        #   total_error += "Your cnv file must not have any empty values\n"

        if process_functions.checkColExist(cnvDF, "ENTREZ_GENE_ID"):
            del cnvDF["ENTREZ_GENE_ID"]
        error = process_functions.validate_genie_identifier(
            identifiers=cnvDF.columns, center=self.center, filename="cnv", col="samples"
        )
        total_error += error
        # cnvDF = cnvDF.fillna('')
        allowed_values = [
            "-2.0",
            "-2",
            "-1.5",
            "-1.0",
            "-1",
            "0.0",
            "0",
            "0.5",
            "1.0",
            "1",
            "1.5",
            "2",
            "2.0",
            "nan",
        ]
        if not all(cnvDF.applymap(lambda x: str(x) in allowed_values).all()):
            total_error += (
                "All values must be NA/blank, -2, -1.5, -1, -0.5, "
                "0, 0.5, 1, 1.5, or 2.\n"
            )
        else:
            cnvDF["HUGO_SYMBOL"] = keepSymbols
            if haveColumn and not nosymbol_check:
                bedSynId = self.genie_config["bed"]
                bed = self.syn.tableQuery(
                    f"select Hugo_Symbol, ID from {bedSynId} "
                    f"where CENTER = '{self.center}'"
                )
                bedDf = bed.asDataFrame()
                cnvDF["remapped"] = cnvDF["HUGO_SYMBOL"].apply(
                    lambda x: validateSymbol(x, bedDf)
                )
                cnvDF = cnvDF[~cnvDF["remapped"].isnull()]

                # Do not allow any duplicated genes after symbols
                # have been remapped
                if sum(cnvDF["remapped"].duplicated()) > 0:
                    duplicated = cnvDF["remapped"].duplicated(keep=False)
                    total_error += (
                        "Your CNA file has duplicated Hugo_Symbols "
                        "(After remapping of genes): {} -> {}.\n".format(
                            ",".join(cnvDF["HUGO_SYMBOL"][duplicated]),
                            ",".join(cnvDF["remapped"][duplicated]),
                        )
                    )
        return (total_error, warning)



import datetime
import logging
import os

import pandas as pd
from synapseclient.core.utils import to_unix_epoch_time

from genie.example_filetype_format import FileTypeFormat
from genie import process_functions

logger = logging.getLogger(__name__)


class sampleRetraction(FileTypeFormat):

    _fileType = "sampleRetraction"

    _process_kwargs = ["newPath", "databaseSynId", "fileSynId"]

    def _get_dataframe(self, filePathList):
        """
        This function by defaults assumes the filePathList is length of 1
        and is a tsv file.  Could change depending on file type.
        """
        filePath = filePathList[0]
        df = pd.read_csv(filePath, header=None)
        return df

    def _validateFilename(self, filePath):
        assert os.path.basename(filePath[0]) == "%s.csv" % self._fileType

    def _process(self, deleteSamplesDf, modifiedOn):
        col = (
            "genieSampleId"
            if self._fileType == "sampleRetraction"
            else "geniePatientId"
        )
        deleteSamplesDf.rename(columns={0: col}, inplace=True)
        modifiedOn = to_unix_epoch_time(
            datetime.datetime.strptime(modifiedOn, "%Y-%m-%dT%H:%M:%S")
        )
        deleteSamplesDf["retractionDate"] = modifiedOn
        deleteSamplesDf["center"] = self.center
        return deleteSamplesDf

    def process_steps(self, deleteSamples, fileSynId, databaseSynId, newPath):
        info = self.syn.get(fileSynId, downloadFile=False)
        deleteSamples = self._process(deleteSamples, info.modifiedOn.split(".")[0])
        process_functions.updateData(
            self.syn,
            databaseSynId,
            deleteSamples,
            self.center,
            filterByColumn="center",
            toDelete=True,
        )
        return newPath



"""Assay information class"""
import os
import yaml

import pandas as pd

from genie.example_filetype_format import FileTypeFormat
from genie import process_functions


class Assayinfo(FileTypeFormat):
    """Assay information file type"""

    _fileType = "assayinfo"

    _process_kwargs = ["newPath", "databaseSynId"]

    _validation_kwargs = ["project_id"]

    def _validateFilename(self, filepath_list):
        """Validate assay information filename"""
        assert os.path.basename(filepath_list[0]) == "assay_information.yaml"

    def process_steps(self, assay_info_df, newPath, databaseSynId):
        """
        Process bed input and update bed database

        Args:
            assay_info_df: Assay information dataframe
            newPath: Path to processed assay information
            databaseSynId: assay information database synapse id

        Returns:
            path to assay information dataframe
        """
        # Must pass in a list
        process_assay_info_df = self._process(assay_info_df)
        process_functions.updateData(
            self.syn,
            databaseSynId,
            process_assay_info_df,
            self.center,
            filterByColumn="CENTER",
            toDelete=True,
        )
        process_assay_info_df.to_csv(newPath, sep="\t", index=False)
        return newPath

    def _process(self, df):
        """
        Process assay_information.yaml. Standardizes SEQ_ASSAY_ID,
        default 10 for gene_padding, and fills in variant_classifications

        Args:
            df: Assay information dataframe

        Returns:
            dataframe: Processed dataframe
        """
        seq_assay_ids = [
            assay.upper().replace("_", "-") for assay in df["SEQ_ASSAY_ID"]
        ]
        df["SEQ_ASSAY_ID"] = seq_assay_ids
        df["SEQ_PIPELINE_ID"] = [
            assay.upper().replace("_", "-") for assay in df["SEQ_PIPELINE_ID"]
        ]
        if process_functions.checkColExist(df, "gene_padding"):
            df["gene_padding"] = df["gene_padding"].fillna(10)
            df["gene_padding"] = df["gene_padding"].astype(int)
        else:
            df["gene_padding"] = 10

        if not process_functions.checkColExist(df, "variant_classifications"):
            df["variant_classifications"] = float("nan")

        df["CENTER"] = self.center
        return df

    def _get_dataframe(self, filepath_list):
        """Take in yaml file, returns dataframe"""
        filepath = filepath_list[0]
        try:
            with open(filepath, "r") as yamlfile:
                # https://github.com/yaml/pyyaml/wiki/PyYAML-yaml.load(input)-Deprecation
                # Must add this because yaml load deprecation
                assay_info_dict = yaml.safe_load(yamlfile)
        except Exception:
            raise ValueError(
                "assay_information.yaml: Can't read in your file. "
                "Please make sure the file is a correctly formatted yaml"
            )
        # assay_info_df = pd.DataFrame(panel_info_dict)
        # assay_info_df = assay_info_df.transpose()
        # assay_info_df['SEQ_ASSAY_ID'] = assay_info_df.index
        # assay_info_df.reset_index(drop=True, inplace=True)
        assay_infodf = pd.DataFrame(assay_info_dict)
        assay_info_transposeddf = assay_infodf.transpose()

        all_panel_info = pd.DataFrame()
        for assay in assay_info_dict:
            assay_specific_info = assay_info_dict[assay]["assay_specific_info"]
            assay_specific_infodf = pd.DataFrame(assay_specific_info)

            intial_seq_id_infodf = assay_info_transposeddf.loc[[assay]]

            # make sure to create a skeleton for the number of seq assay ids
            # in the seq pipeline
            seq_assay_id_infodf = pd.concat(
                [intial_seq_id_infodf] * len(assay_specific_info)
            )
            seq_assay_id_infodf.reset_index(drop=True, inplace=True)
            assay_finaldf = pd.concat(
                [assay_specific_infodf, seq_assay_id_infodf], axis=1
            )
            del assay_finaldf["assay_specific_info"]
            # Transform values containing lists to string concatenated values
            columns_containing_lists = [
                "variant_classifications",
                "alteration_types",
                "preservation_technique",
                "coverage",
            ]

            for col in columns_containing_lists:
                if assay_finaldf.get(col) is not None:
                    assay_finaldf[col] = [";".join(row) for row in assay_finaldf[col]]
            assay_finaldf["SEQ_PIPELINE_ID"] = assay
            all_panel_info = pd.concat([all_panel_info, assay_finaldf])
        return all_panel_info

    def _validate(self, assay_info_df, project_id):
        """
        Validates the values of assay information file

        Args:
            assay_info_df: assay information dataframe

        Returns:
            tuple: error and warning
        """

        total_error = ""
        warning = ""

        if process_functions.checkColExist(assay_info_df, "SEQ_ASSAY_ID"):
            all_seq_assays = (
                assay_info_df.SEQ_ASSAY_ID.replace({"_": "-"}, regex=True)
                .str.upper()
                .unique()
            )
            if not all([assay.startswith(self.center) for assay in all_seq_assays]):
                total_error += (
                    "Assay_information.yaml: Please make sure all your "
                    "SEQ_ASSAY_IDs start with your center abbreviation.\n"
                )
            db_to_syn_map_df = process_functions.get_synid_database_mappingdf(
                self.syn, project_id
            )
            sample_synid = process_functions.getDatabaseSynId(
                self.syn, "sample", databaseToSynIdMappingDf=db_to_syn_map_df
            )
            uniq_seq_df = process_functions.get_syntabledf(
                self.syn,
                f"select distinct(SEQ_ASSAY_ID) as seq from {sample_synid} "
                f"where CENTER = '{self.center}'",
            )
            # These are all the SEQ_ASSAY_IDs that are in the clinical database
            # but not in the assay_information file
            missing_seqs = uniq_seq_df["seq"][
                ~uniq_seq_df["seq"]
                .replace({"_": "-"}, regex=True)
                .str.upper()
                .isin(all_seq_assays)
            ]
            missing_seqs_str = ", ".join(missing_seqs)
            if missing_seqs.to_list():
                total_error += (
                    "Assay_information.yaml: You are missing SEQ_ASSAY_IDs: "
                    f"{missing_seqs_str}\n"
                )

        else:
            total_error += "Assay_information.yaml: Must have SEQ_ASSAY_ID column.\n"

        read_group_dict = process_functions.get_gdc_data_dictionary("read_group")
        read_group_headers = read_group_dict["properties"]
        warn, error = process_functions.check_col_and_values(
            assay_info_df,
            "is_paired_end",
            [True, False],
            filename="Assay_information.yaml",
            required=True,
        )
        warning += warn
        total_error += error

        warn, error = process_functions.check_col_and_values(
            assay_info_df,
            "library_selection",
            read_group_headers["library_selection"]["enum"],
            filename="Assay_information.yaml",
            required=True,
        )
        warning += warn
        total_error += error

        warn, error = process_functions.check_col_and_values(
            assay_info_df,
            "library_strategy",
            read_group_headers["library_strategy"]["enum"],
            filename="Assay_information.yaml",
            required=True,
        )
        warning += warn
        total_error += error

        warn, error = process_functions.check_col_and_values(
            assay_info_df,
            "platform",
            read_group_headers["platform"]["enum"],
            filename="Assay_information.yaml",
            required=True,
        )
        warning += warn
        total_error += error

        instrument_model = read_group_headers["instrument_model"]["enum"]
        instrument_model.extend(["Illumina NovaSeq 6000", None])
        warn, error = process_functions.check_col_and_values(
            assay_info_df,
            "instrument_model",
            instrument_model,
            filename="Assay_information.yaml",
            required=True,
        )
        warning += warn
        total_error += error

        # target_capture_kit = read_group_headers['target_capture_kit']['enum']
        # warn, error = process_functions.check_col_and_values(
        #     assay_info_df,
        #     'target_capture_kit',
        #     target_capture_kit,
        #     filename="Assay_information.yaml",
        #     required=True)
        # warning += warn
        # total_error += error

        if not process_functions.checkColExist(assay_info_df, "target_capture_kit"):
            total_error += (
                "Assay_information.yaml: " "Must have target_capture_kit column.\n"
            )

        variant_classes = [
            "Splice_Site",
            "Nonsense_Mutation",
            "Frame_Shift_Del",
            "Frame_Shift_Ins",
            "Nonstop_Mutation",
            "Translation_Start_Site",
            "In_Frame_Ins",
            "In_Frame_Del",
            "Missense_Mutation",
            "Intron",
            "Splice_Region",
            "Silent",
            "RNA",
            "5'UTR",
            "3'UTR",
            "IGR",
            "5'Flank",
            "3'Flank",
            None,
        ]
        warn, error = process_functions.check_col_and_values(
            assay_info_df,
            "variant_classifications",
            variant_classes,
            filename="Assay_information.yaml",
            na_allowed=True,
            sep=";",
        )
        warning += warn
        total_error += error

        if process_functions.checkColExist(assay_info_df, "read_length"):
            if not all(
                [
                    process_functions.checkInt(i)
                    for i in assay_info_df["read_length"]
                    if i is not None and not pd.isnull(i)
                ]
            ):
                total_error += (
                    "Assay_information.yaml: "
                    "Please double check your read_length.  "
                    "It must be an integer or null.\n"
                )
        else:
            total_error += "Assay_information.yaml: " "Must have read_length column.\n"

        if process_functions.checkColExist(assay_info_df, "number_of_genes"):
            if not all(
                [
                    process_functions.checkInt(i)
                    for i in assay_info_df["number_of_genes"]
                ]
            ):
                total_error += (
                    "Assay_information.yaml: "
                    "Please double check your number_of_genes. "
                    "It must be an integer.\n"
                )
        else:
            total_error += (
                "Assay_information.yaml: " "Must have number_of_genes column.\n"
            )

        if process_functions.checkColExist(assay_info_df, "gene_padding"):
            if not all(
                [
                    process_functions.checkInt(i)
                    for i in assay_info_df["gene_padding"]
                    if i is not None and not pd.isnull(i)
                ]
            ):
                total_error += (
                    "Assay_information.yaml: "
                    "Please double check your gene_padding. "
                    "It must be an integer or blank.\n"
                )
        else:
            warning += (
                "Assay_information.yaml: "
                "gene_padding is by default 10 if not specified.\n"
            )

        warn, error = process_functions.check_col_and_values(
            assay_info_df,
            "calling_strategy",
            ["tumor_only", "tumor_normal", "plasma_normal"],
            filename="Assay_information.yaml",
            required=True,
        )
        warning += warn
        total_error += error

        if process_functions.checkColExist(assay_info_df, "specimen_tumor_cellularity"):
            if not all(
                [
                    i.startswith(">") and i.endswith("%")
                    for i in assay_info_df["specimen_tumor_cellularity"]
                ]
            ):
                total_error += (
                    "Assay_information.yaml: "
                    "Please double check your specimen_tumor_cellularity. "
                    "It must in this format >(num)%. ie. >10%\n"
                )
        else:
            total_error += (
                "Assay_information.yaml: "
                "Must have specimen_tumor_cellularity column.\n"
            )

        alteration_types = [
            "snv",
            "small_indels",
            "gene_level_cna",
            "intragenic_cna",
            "structural_variants",
        ]
        warn, error = process_functions.check_col_and_values(
            assay_info_df,
            "alteration_types",
            alteration_types,
            filename="Assay_information.yaml",
            required=True,
            sep=";",
        )
        warning += warn
        total_error += error

        preservation_technique = ["FFPE", "fresh_frozen", "NA"]
        warn, error = process_functions.check_col_and_values(
            assay_info_df,
            "preservation_technique",
            preservation_technique,
            filename="Assay_information.yaml",
            required=True,
            sep=";",
        )
        warning += warn
        total_error += error

        coverage = ["hotspot_regions", "coding_exons", "introns", "promoters"]
        warn, error = process_functions.check_col_and_values(
            assay_info_df,
            "coverage",
            coverage,
            filename="Assay_information.yaml",
            required=True,
            sep=";",
        )
        warning += warn
        total_error += error

        return total_error, warning



"""Tests database to staging functions"""
import os
from unittest import mock
from unittest.mock import patch

import pandas as pd
import synapseclient

from genie import database_to_staging, process_functions

SYN = synapseclient.Synapse()
FILEVIEW_SYNID = "syn12345"
GENIE_VERSION = "vTEST"
CONSORTIUM_SYNID = "syn2222"


class Tablequerydf:
    """tableQuery.asDataFrame() class"""

    def __init__(self, df):
        self.df = df

    def asDataFrame(self):
        return self.df


def test_store_gene_panel_files():
    data_gene_panel = pd.DataFrame({"mutations": ["PANEL1"]})
    gene_paneldf = pd.DataFrame({"id": ["syn3333"]})

    with mock.patch.object(
        SYN, "tableQuery", return_value=Tablequerydf(gene_paneldf)
    ) as patch_syn_table_query, mock.patch.object(
        database_to_staging, "store_file", return_value=synapseclient.Entity()
    ) as patch_storefile, mock.patch.object(
        SYN,
        "get",
        return_value=synapseclient.Entity(path="/foo/bar/PANEL1.txt", versionNumber=2),
    ) as patch_syn_get, mock.patch.object(
        os, "rename"
    ) as patch_os_rename:

        database_to_staging.store_gene_panel_files(
            SYN,
            FILEVIEW_SYNID,
            GENIE_VERSION,
            data_gene_panel,
            CONSORTIUM_SYNID,
            ["TEST"],
        )

        patch_syn_table_query.assert_called_once_with(
            "select id from %s where cBioFileFormat = 'genePanel' "
            "and fileStage = 'staging' and "
            "name not in ('data_gene_panel_TEST.txt')" % FILEVIEW_SYNID
        )

        patch_storefile.assert_called_once_with(
            SYN,
            os.path.join(database_to_staging.GENIE_RELEASE_DIR, "PANEL1_vTEST.txt"),
            parent=CONSORTIUM_SYNID,
            genieVersion=GENIE_VERSION,
            name="PANEL1.txt",
            cBioFileFormat="genePanel",
            used="syn3333.2",
        )

        patch_syn_get.assert_called_once_with("syn3333")
        patch_os_rename.assert_called_once_with(
            "/foo/bar/PANEL1.txt",
            os.path.join(database_to_staging.GENIE_RELEASE_DIR, "PANEL1_vTEST.txt"),
        )


def test_store_assay_info_files():
    """Tests storing of assay information file"""
    assay_infodf = pd.DataFrame({"library_strategy": ["WXS"], "SEQ_ASSAY_ID": ["A"]})
    clinicaldf = pd.DataFrame({"SEQ_ASSAY_ID": ["A"]})
    database_to_staging.GENIE_RELEASE_DIR = "./"
    path = os.path.join(
        database_to_staging.GENIE_RELEASE_DIR, "assay_information_vTEST.txt"
    )
    with patch.object(
        SYN, "create_snapshot_version", return_value=2
    ) as patch_create_version, patch.object(
        process_functions, "get_syntabledf", return_value=assay_infodf
    ) as patch_table_query, patch.object(
        database_to_staging, "store_file", return_value=synapseclient.Entity()
    ) as patch_storefile:
        wes_ids = database_to_staging.store_assay_info_files(
            SYN, GENIE_VERSION, FILEVIEW_SYNID, clinicaldf, CONSORTIUM_SYNID
        )
        patch_create_version.assert_called_once_with(
            FILEVIEW_SYNID, comment=GENIE_VERSION
        )
        patch_table_query.assert_called_once_with(
            SYN, f"select * from {FILEVIEW_SYNID} where SEQ_ASSAY_ID in ('A')"
        )
        patch_storefile.assert_called_once_with(
            SYN,
            path,
            parent=CONSORTIUM_SYNID,
            genieVersion=GENIE_VERSION,
            name="assay_information.txt",
            used=f"{FILEVIEW_SYNID}.2",
        )
        assert wes_ids == ["A"]



"""Write invalid reasons"""
import logging
import os

import pandas as pd
import synapseclient  # lgtm [py/import-and-import-from]
from synapseclient import Synapse

from . import process_functions

logger = logging.getLogger(__name__)


def write(syn: Synapse, center_mapping_synid: str, error_tracker_synid: str):
    """Write center errors to a file

    Args:
        syn: Synapse connection
        center_mapping_df: Center mapping dataframe
        error_tracker_synid: Error tracking synapse id

    """
    center_mapping_df = process_functions.get_syntabledf(
        syn=syn,
        query_string=f"SELECT * FROM {center_mapping_synid} where release is true",
    )
    center_errors = get_center_invalid_errors(syn, error_tracker_synid)
    for center in center_mapping_df["center"]:
        logger.info(center)
        staging_synid = center_mapping_df["stagingSynId"][
            center_mapping_df["center"] == center
        ][0]
        with open(center + "_errors.txt", "w") as errorfile:
            if center not in center_errors:
                errorfile.write("No errors!")
            else:
                errorfile.write(center_errors[center])

        ent = synapseclient.File(center + "_errors.txt", parentId=staging_synid)
        syn.store(ent)
        os.remove(center + "_errors.txt")


def _combine_center_file_errors(syn: Synapse, center_errorsdf: pd.DataFrame) -> str:
    """Combine all center errors into one printable string

    Args:
        syn: Synapse connection
        center_errorsdf: Center errors dataframe

    Returns:
        Center errors in a pretty formatted string

    """
    center_errors = ""
    for _, row in center_errorsdf.iterrows():
        ent = syn.get(row["id"], downloadFile=False)
        file_errors = row["errors"].replace("|", "\n")
        error_text = f"\t{ent.name} ({ent.id}):\n\n{file_errors}\n\n"
        center_errors += error_text
    return center_errors


def get_center_invalid_errors(syn: Synapse, error_tracker_synid: str) -> dict:
    """Get all invalid errors per center

    Args:
        syn: Synapse connection
        error_tracker_synid: Synapse id of invalid error database table

    Returns:
        dict: center - file error string

    """
    error_tracker = syn.tableQuery(f"SELECT * FROM {error_tracker_synid}")
    error_trackerdf = error_tracker.asDataFrame()
    center_errorsdf = error_trackerdf.groupby("center")
    center_error_map = {}
    for center, df in center_errorsdf:
        center_error_map[center] = _combine_center_file_errors(syn, df)
    return center_error_map



"""Processing functions that are used in the GENIE pipeline"""
import datetime
import json
import logging
import os
import requests
from requests.adapters import HTTPAdapter
from requests.packages.urllib3.util.retry import Retry
import tempfile
import time
from typing import List

import ast
from Crypto.Cipher import PKCS1_OAEP
from Crypto.PublicKey import RSA
import pandas as pd
import synapseclient  # lgtm [py/import-and-import-from]
from synapseclient import Synapse

# try:
#   from urllib.request import urlopen
# except ImportError:
#   from urllib2 import urlopen
# Ignore SettingWithCopyWarning warning
pd.options.mode.chained_assignment = None

logger = logging.getLogger(__name__)

SCRIPT_DIR = os.path.dirname(os.path.abspath(__file__))


# try:
#   from urlparse import urlparse
# except ImportError:
#   from urllib.parse import urlparse

# Create merged dictionary of remapped genes
# def createDict(invalidated_genes):
#     toRemapRemove = {}
#     invalidated_genes = pd.Series(invalidated_genes)
#     for i in invalidated_genes[invalidated_genes!=True]:
#         toRemapRemove.update(i)
#     return(toRemapRemove)

# Validate genes
# def hgncRestCall(path):
#     """
#     This function does the rest call to the genenames website

#     :params path:     The gene symbol url path to add to the base uri

#     :returns:         If the symbol exists, returns True and the corrected symbol, otherwise returns False and None.
#     """
#     headers = {'Accept': 'application/json',}

#     uri = 'http://rest.genenames.org'

#     target = urlparse(uri+path)
#     method = 'GET'
#     body = ''
#     h = http.Http()
#     response, content = h.request(target.geturl(),
#                                   method,
#                                   body,
#                                   headers)
#     if response['status'] == '200':
#         data = json.loads(content)
#         if len(data['response']['docs']) == 0:
#             return(False, [None])
#         else:
#             mapped = [symbol['symbol'] for symbol in data['response']['docs']]
#             return(True, mapped)
#     else:
#         return(False, [None])

# Validation of gene names
# def validateSymbol(gene, returnMapping=False):
#     """
#     This function does validation of symbols

#     :params gene:               Gene symbol
#     :params returnMapping:      Return mapping of old gene to new gene

#     :returns:                   Check if the provided gene name is a correct symbol and print out genes
#                                 that need to be remapped or can't be mapped to anything
#     """
#     path = '/fetch/symbol/%s' %  gene
#     verified, symbol = hgncRestCall(path)
#     if not verified:
#         path = '/fetch/prev_symbol/%s' %  gene
#         verified, symbol = hgncRestCall(path)
#     if not verified:
#         path = '/fetch/alias_symbol/%s' %  gene
#         verified, symbol = hgncRestCall(path)
#     if gene in symbol:
#         return(True)
#     else:
#         if symbol[0] is None:
#             #logger.error("%s cannot be remapped. Please correct." % gene)
#             logger.warning("%s cannot be remapped. These rows will have an empty gene symbol" % gene)
#         else:
#             #if "MLL4", then the HUGO symbol should be KMT2D and KMT2B
#             if len(symbol) > 1:
#                 #logger.error("%s can be mapped to different symbols: %s. Please correct." % (gene, ", ".join(symbol)))
#                 logger.warning("%s can be mapped to different symbols: %s. Please correct or it will be removed." % (gene, ", ".join(symbol)))
#             else:
#                 logger.info("%s will be remapped to %s" % (gene, symbol[0]))
#                 if returnMapping:
#                     return({gene: symbol[0]})
#                 else:
#                     return(True)
#         if returnMapping:
#             return({gene: float('nan')})
#         else:
#             return(False)
#         #return(True)

# # Remap and remove genes in dataframes given a column
# def remapGenes(invalidated_genes, DF, col,isBedFile=False):
#     nonmapped = []
#     gene_dict = createDict(invalidated_genes)
#     for key in gene_dict:
#         value = gene_dict[key]
#         if isBedFile:
#         #     nonmapped.append(key)
#         #     DF = DF[DF[col] != key]
#         # else:
#             DF[col][DF[col] == key] = value
#     return(DF, nonmapped)


def retry_get_url(url):
    """
    Implement retry logic when getting urls.
    Timesout at 3 seconds, retries 5 times.

    Args:
        url:  Http or https url

    Returns:
        requests.get()
    """
    s = requests.Session()
    retries = Retry(total=5, backoff_factor=1)
    s.mount("http://", HTTPAdapter(max_retries=retries))
    s.mount("https://", HTTPAdapter(max_retries=retries))
    response = s.get(url, timeout=3)
    return response


def checkUrl(url):
    """
    Check if URL link is live

    Args:
        url: web URL
    """
    temp = retry_get_url(url)
    assert temp.status_code == 200, "%s site is down" % url


def checkColExist(DF, key):
    """
    This function checks if the column exists in a dataframe

    Args:
        DF: pandas dataframe
        key: Expected column header name

    Returns:
        bool:  True if column exists
    """
    result = False if DF.get(key) is None else True
    return result


def validate_genie_identifier(
    identifiers: pd.Series, center: str, filename: str, col: str
) -> str:
    """Validate GENIE sample and patient ids.

    Args:
        identifiers (pd.Series): Array of GENIE identifiers
        center (str): GENIE center name
        filename (str): name of file
        col (str): Column with identifiers

    return:
        str: Errors
    """
    total_error = ""
    if not all(identifiers.str.startswith(f"GENIE-{center}")):
        total_error = total_error + (
            f"{filename}: {col} must start with GENIE-{center}\n"
        )
    if any(identifiers.str.len() >= 50):
        total_error = total_error + (
            f"{filename}: {col} must have less than 50 characters.\n"
        )
    return total_error


# def get_oncotree_codes(oncotree_url):
#     '''
#     Deprecated
#     Gets the oncotree data from the specified url.

#     Args:
#         oncotree_url: Oncotree url

#     Returns:
#         dataframe: oncotree codes
#     '''
#     # PATTERN = re.compile('([A-Za-z\' ,-/]*) \\(([A-Za-z_]*)\\)')
#     PATTERN = re.compile('.*[(](.*)[)]')
#     # with requests.get(oncotree_url) as oncotreeUrl:
#     #   oncotree = oncotreeUrl.text.split("\n")
#     oncotreeUrl = retry_get_url(oncotree_url)
#     oncotree = oncotreeUrl.text.split("\n")

#     # oncotree = urlopen(oncotree_url).read().split('\n')
#     allCodes = []
#     for row in oncotree[:-1]:
#         data = row.split("\t")
#         allCodes.extend([
#             PATTERN.match(i).group(1)
#             for i in data[0:5] if PATTERN.match(i)])
#     codes = pd.DataFrame({"ONCOTREE_CODE": list(set(allCodes))})
#     return(codes)


def lookup_dataframe_value(df, col, query):
    """
    Look up dataframe value given query and column

    Args:
        df: dataframe
        col: column with value to return
        query: Query for specific column

    Returns:
        value
    """
    query = df.query(query)
    query_val = query[col].iloc[0]
    return query_val


def get_syntabledf(syn, query_string):
    """
    Get dataframe from table query

    Args:
        syn: Synapse object
        query_string: Table query

    Returns:
        pandas dataframe with query results
    """
    table = syn.tableQuery(query_string)
    tabledf = table.asDataFrame()
    return tabledf


def get_synid_database_mappingdf(syn, project_id):
    """
    Get database to synapse id mapping dataframe

    Args:
        syn: Synapse object
        project_id: Synapse Project ID with a 'dbMapping' annotation.

    Returns:
        database to synapse id mapping dataframe
    """

    project = syn.get(project_id)
    database_mapping_synid = project.annotations["dbMapping"][0]
    database_map_query = "SELECT * FROM {}".format(database_mapping_synid)
    mappingdf = get_syntabledf(syn, database_map_query)
    return mappingdf


def getDatabaseSynId(syn, tableName, project_id=None, databaseToSynIdMappingDf=None):
    """
    Get database synapse id from database to synapse id mapping table

    Args:
        syn: Synapse object
        project_id: Synapse Project ID with a database mapping table.
        tableName: Name of synapse table
        databaseToSynIdMappingDf: Avoid calling rest call to download table
                                  if the mapping table is already downloaded

    Returns:
        str:  Synapse id of wanted database
    """
    if databaseToSynIdMappingDf is None:
        databaseToSynIdMappingDf = get_synid_database_mappingdf(
            syn, project_id=project_id
        )

    synId = lookup_dataframe_value(
        databaseToSynIdMappingDf, "Id", 'Database == "{}"'.format(tableName)
    )
    return synId


def rmFiles(folderPath, recursive=True):
    """
    Convenience function to remove all files in dir

    Args:
        folderPath: Path to folder
        recursive:  Removes all files recursively
    """
    for dirPath, dirNames, filePaths in os.walk(folderPath):
        for filePath in filePaths:
            os.unlink(os.path.join(dirPath, filePath))
        if not recursive:
            break


def removeStringFloat(string):
    """
    remove string float in tsv file

    Args:
        string: tsv file in string format

    Return:
        string: string with float removed
    """
    string = string.replace(".0\t", "\t")
    string = string.replace(".0\n", "\n")
    return string


def removePandasDfFloat(df, header=True):
    """
    Remove decimal for integers due to pandas

    Args:
        df:  Pandas dataframe

    Return:
        str: tsv in text
    """
    if header:
        text = df.to_csv(sep="\t", index=False)
    else:
        text = df.to_csv(sep="\t", index=False, header=None)

    text = removeStringFloat(text)
    return text


def removeFloat(df):
    """
    Need to remove this function
    as it calls another function
    """
    # text = df.to_csv(sep="\t",index=False)
    # text = text.replace(".0\t","\t")
    # text = text.replace(".0\n","\n")
    text = removePandasDfFloat(df)
    return text


def checkGenieId(ID, center):
    """
    Checks if GENIE ID is labelled correctly
    and reformats the GENIE ID

    Args:
        ID: string
        center: GENIE center

    Return:
        str: Formatted GENIE ID string
    """
    if str(ID).startswith("%s-" % center):
        return "GENIE-%s" % str(ID)
    elif not str(ID).startswith("GENIE-%s-" % center):
        return "GENIE-%s-%s" % (center, str(ID))
    else:
        return str(ID)


def storeFile(
    syn,
    fileName,
    parentId,
    center,
    fileFormat,
    dataSubType,
    platform=None,
    cBioFileFormat=None,
    used=None,
):
    """
    # Storing Files along with annotations
    """
    logger.info("STORING FILES")
    fileEnt = synapseclient.File(fileName, parent=parentId)
    fileEnt.center = center
    fileEnt.species = "Human"
    fileEnt.consortium = "GENIE"
    fileEnt.dataType = "genomicVariants"
    fileEnt.fundingAgency = "AACR"
    fileEnt.assay = "targetGeneSeq"
    fileEnt.fileFormat = fileFormat
    fileEnt.dataSubType = dataSubType
    fileEnt.fileStage = "staging"
    fileEnt.platform = platform
    if platform is not None:
        fileEnt.platform = platform
    if cBioFileFormat is not None:
        fileEnt.cBioFileFormat = cBioFileFormat
    ent = syn.store(fileEnt, used=used)
    return ent


def seqDateFilter(clinicalDf, processingDate, days):
    """
    SEQ_DATE filter
    SEQ_DATE - Clinical data (6 and 12 as parameters)
    Jan-2017 , given processing date (today) ->
        staging release (processing date - Jan-2017 < 6 months)
    July-2016 , given processing date (today) ->
        consortium release (processing date - July-2016 between
        6 months - 12 months)

    """
    copyClinicalDf = clinicalDf.copy()
    # copyClinicalDf['SEQ_DATE'][copyClinicalDf['SEQ_DATE'].astype(str) == '999'] = "Jan-1988"
    # copyClinicalDf['SEQ_DATE'][copyClinicalDf['SEQ_DATE'].astype(str) == '999.0'] = "Jan-1988"
    if not isinstance(processingDate, datetime.datetime):
        processingDate = datetime.datetime.strptime(processingDate, "%b-%Y")
    # Remove this null statement after clinical files have been re-validated
    # copyClinicalDf['SEQ_DATE'][copyClinicalDf['SEQ_DATE'].isnull()] = "Jan-1900"
    copyClinicalDf["SEQ_DATE"][copyClinicalDf["SEQ_DATE"] == "Release"] = "Jan-1900"
    # clinicalDf['SEQ_DATE'][clinicalDf['SEQ_DATE'] == '999'] = "Jan-1988"
    dates = copyClinicalDf["SEQ_DATE"].apply(
        lambda date: datetime.datetime.strptime(date, "%b-%Y")
    )
    keep = processingDate - dates > datetime.timedelta(days)
    keepSamples = copyClinicalDf["SAMPLE_ID"][~keep]
    # copyClinicalDf.SEQ_DATE[keep].unique()
    return keepSamples


def addClinicalHeaders(
    clinicalDf, mapping, patientCols, sampleCols, samplePath, patientPath
):
    """
    Add clinical file headers

    Args:
        clinicalDf: clinical dataframe
        mapping: mapping dataframe, maps clinical columns to
                 labels and descriptions
        patientCols: list of patient columns
        sampleCols: list of sample columns
        samplePath: clinical sample path
        patientPath: clinical patient path
    """
    patientLabels = [
        str(mapping["labels"][mapping["cbio"] == i].values[0]) for i in patientCols
    ]
    sampleLabels = [
        str(mapping["labels"][mapping["cbio"] == i].values[0]) for i in sampleCols
    ]
    patientDesc = [
        str(mapping["description"][mapping["cbio"] == i].values[0]) for i in patientCols
    ]
    sampleDesc = [
        str(mapping["description"][mapping["cbio"] == i].values[0]) for i in sampleCols
    ]
    patientType = [
        str(mapping["colType"][mapping["cbio"] == i].values[0]) for i in patientCols
    ]
    sampleType = [
        str(mapping["colType"][mapping["cbio"] == i].values[0]) for i in sampleCols
    ]

    with open(patientPath, "w+") as patientFile:
        patientFile.write("#%s\n" % "\t".join(patientLabels))
        patientFile.write("#%s\n" % "\t".join(patientDesc))
        patientFile.write("#%s\n" % "\t".join(patientType))
        patientFile.write("#%s\n" % "\t".join(["1"] * len(patientLabels)))
        text = removeFloat(clinicalDf[patientCols].drop_duplicates("PATIENT_ID"))
        patientFile.write(text)
    with open(samplePath, "w+") as sampleFile:
        sampleFile.write("#%s\n" % "\t".join(sampleLabels))
        sampleFile.write("#%s\n" % "\t".join(sampleDesc))
        sampleFile.write("#%s\n" % "\t".join(sampleType))
        sampleFile.write("#%s\n" % "\t".join(["1"] * len(sampleLabels)))
        text = removeFloat(clinicalDf[sampleCols].drop_duplicates("SAMPLE_ID"))
        sampleFile.write(text)


########################################################################
# CENTER ANONYMIZING
########################################################################


# def center_anon(filePath, anonymizeCenterDf):
#     '''
#     Deprecated
#     '''
#     with open(filePath, "r") as datafile:
#         text = datafile.read()
#     for center in anonymizeCenterDf['center']:
#         newCenter = anonymizeCenterDf['newCenter'][
#             anonymizeCenterDf['center'] == center].values[0]
#         text = re.sub("\t%s\t" % center, "\t%s\t" % newCenter, text)
#         text = re.sub("GENIE-%s-" % center, "GENIE-%s-" % newCenter, text)
#     with open(filePath, "w") as datafile:
#         datafile.write(text)


# def center_convert_back(filePath, anonymizeCenterDf):
#     '''
#     Deprecated
#     '''
#     with open(filePath, "r") as datafile:
#         text = datafile.read()
#     for center in anonymizeCenterDf['center']:
#         newCenter = anonymizeCenterDf['newCenter'][
#             anonymizeCenterDf['center'] == center].values[0]
#         text = re.sub("\t%s\t" % newCenter, "\t%s\t" % center, text)
#         text = re.sub("GENIE-%s-" % newCenter, "GENIE-%s-" % center, text)
#     with open(filePath, "w") as datafile:
#         datafile.write(text)

##############################################################################
# UPDATING DATABASE
##############################################################################


def _check_valid_df(df, col):
    """
    Checking if variable is a pandas dataframe and column specified exist

    Args:
        df: Pandas dataframe
        col: Column name
    """
    if not isinstance(df, pd.DataFrame):
        raise ValueError("Must pass in pandas dataframe")
    if df.get(col) is None:
        raise ValueError("'{}' column must exist in dataframe".format(col))


def _get_left_diff_df(left, right, checkby):
    """
    Subset the dataframe based on 'checkby' by taking values in the left df
    that arent in the right df

    Args:
        left: Dataframe
        right: Dataframe
        checkby: Column of values to compare

    Return:
        Dataframe: Subset of dataframe from left that don't exist in the right
    """
    _check_valid_df(left, checkby)
    _check_valid_df(right, checkby)
    diffdf = left[~left[checkby].isin(right[checkby])]
    return diffdf


def _get_left_union_df(left, right, checkby):
    """
    Subset the dataframe based on 'checkby' by taking the union of
    values in the left df with the right df

    Args:
        left: Dataframe
        right: Dataframe
        checkby: Column of values to compare

    Return:
        Dataframe: Subset of dataframe from left that also exist in the right
    """
    _check_valid_df(left, checkby)
    _check_valid_df(right, checkby)
    uniondf = left[left[checkby].isin(right[checkby])]
    return uniondf


def _append_rows(new_datasetdf, databasedf, checkby):
    """
    Compares the dataset from the database and determines which rows to
    append from the dataset

    Args:
        new_datasetdf: Input data dataframe
        databasedf: Existing data dataframe
        checkby: Column of values to compare

    Return:
        Dataframe: Dataframe of rows to append
    """
    databasedf.fillna("", inplace=True)
    new_datasetdf.fillna("", inplace=True)

    appenddf = _get_left_diff_df(new_datasetdf, databasedf, checkby)
    if not appenddf.empty:
        logger.info("Adding Rows")
    else:
        logger.info("No new rows")
    del appenddf[checkby]
    appenddf.reset_index(drop=True, inplace=True)
    return appenddf


def _delete_rows(new_datasetdf, databasedf, checkby):
    """
    Compares the dataset from the database and determines which rows to
    delete from the dataset

    Args:
        new_datasetdf: Input data dataframe
        databasedf: Existing data dataframe
        checkby: Column of values to compare

    Return:
        Dataframe: Dataframe of rows to delete
    """

    databasedf.fillna("", inplace=True)
    new_datasetdf.fillna("", inplace=True)
    # If the new dataset is empty, delete everything in the database
    deletedf = _get_left_diff_df(databasedf, new_datasetdf, checkby)
    if not deletedf.empty:
        logger.info("Deleting Rows")
        delete_rowid_version = pd.DataFrame(
            [[rowid.split("_")[0], rowid.split("_")[1]] for rowid in deletedf.index]
        )
        delete_rowid_version.reset_index(drop=True, inplace=True)
    else:
        delete_rowid_version = pd.DataFrame()
        logger.info("No deleted rows")

    # del deletedf[checkby]
    return delete_rowid_version


def _create_update_rowsdf(updating_databasedf, updatesetdf, rowids, differentrows):
    """
    Create the update dataset dataframe

    Args:
        updating_databasedf: Update database dataframe
        updatesetdf:  Update dataset dataframe
        rowids: rowids of the database (Synapse ROW_ID, ROW_VERSION)
        differentrows: vector of booleans for rows that need to be updated
                       True for update, False for not

    Returns:
        dataframe: Update dataframe
    """
    if sum(differentrows) > 0:
        updating_databasedf.loc[differentrows] = updatesetdf.loc[differentrows]
        toupdatedf = updating_databasedf.loc[differentrows]
        logger.info("Updating rows")
        rowid_version = pd.DataFrame(
            [
                [rowid.split("_")[0], rowid.split("_")[1]]
                for rowid, row in zip(rowids, differentrows)
                if row
            ]
        )
        toupdatedf["ROW_ID"] = rowid_version[0].values
        toupdatedf["ROW_VERSION"] = rowid_version[1].values
        toupdatedf.reset_index(drop=True, inplace=True)
    else:
        toupdatedf = pd.DataFrame()
        logger.info("No updated rows")
    return toupdatedf


def _update_rows(new_datasetdf, databasedf, checkby):
    """
    Compares the dataset from the database and determines which rows to
    update from the dataset

    Args:
        new_datasetdf: Input data dataframe
        databasedf: Existing data dataframe
        checkby: Column of values to compare

    Return:
        Dataframe: Dataframe of rows to update
    """
    # initial_database = databasedf.copy()
    databasedf.fillna("", inplace=True)
    new_datasetdf.fillna("", inplace=True)
    updatesetdf = _get_left_union_df(new_datasetdf, databasedf, checkby)
    updating_databasedf = _get_left_union_df(databasedf, new_datasetdf, checkby)

    # If you input the exact same dataframe theres nothing to update
    # must save row version and ids for later
    rowids = updating_databasedf.index.values
    # Set index values to be 'checkby' values
    updatesetdf.index = updatesetdf[checkby]
    updating_databasedf.index = updating_databasedf[checkby]
    del updatesetdf[checkby]
    del updating_databasedf[checkby]

    # Remove duplicated index values
    updatesetdf = updatesetdf[~updatesetdf.index.duplicated()]
    # Reorder dataset index
    updatesetdf = updatesetdf.loc[updating_databasedf.index]
    # Index comparison
    differences = updatesetdf != updating_databasedf
    differentrows = differences.apply(sum, axis=1) > 0

    toupdatedf = _create_update_rowsdf(
        updating_databasedf, updatesetdf, rowids, differentrows
    )

    return toupdatedf


def updateData(
    syn: Synapse,
    databaseSynId: str,
    newData: pd.DataFrame,
    filterBy: str,
    filterByColumn: str = "CENTER",
    col: List[str] = None,
    toDelete: bool = False,
) -> None:
    """Update Synapse table given a new dataframe

    Args:
        syn (Synapse): Synapse connection
        databaseSynId (str): Synapse Id of Synapse Table
        newData (pd.DataFrame): New data in a dataframe
        filterBy (str): Value to filter new data by
        filterByColumn (str, optional): Column to filter values by. Defaults to "CENTER".
        col (List[str], optional): List of columns to ingest. Defaults to None.
        toDelete (bool, optional): Delete rows given the primary key. Defaults to False.
    """
    databaseEnt = syn.get(databaseSynId)
    database = syn.tableQuery(
        f"SELECT * FROM {databaseSynId} where {filterByColumn} ='{filterBy}'"
    )
    database = database.asDataFrame()
    db_cols = set(database.columns)
    if col is not None:
        new_data_cols = set(col)
        # Make sure columns from file exists in database columns
        use_cols = db_cols.intersection(new_data_cols)
        # No need to fail, because there is bound to be at least one
        # column that will exist in the database
        database = database[list(use_cols)]
    else:
        newData = newData[database.columns]
    updateDatabase(
        syn, database, newData, databaseSynId, databaseEnt.primaryKey, toDelete
    )


def updateDatabase(
    syn, database, new_dataset, database_synid, primary_key_cols, to_delete=False
):
    """
    Updates synapse tables by a row identifier with another
    dataset that has the same number and order of columns

    Args:
        syn: Synapse object
        database: The synapse table (pandas dataframe)
        new_dataset: New dataset (pandas dataframe)
        databaseSynId: Synapse Id of the database table
        uniqueKeyCols: Column(s) that make up the unique key
        toDelete: Delete rows, Defaults to False

    Returns:
        Nothing
    """
    primary_key = "UNIQUE_KEY"
    database = database.fillna("")
    orig_database_cols = database.columns
    col_order = ["ROW_ID", "ROW_VERSION"]
    col_order.extend(orig_database_cols.tolist())
    new_dataset = new_dataset.fillna("")
    # Columns must be in the same order
    new_dataset = new_dataset[orig_database_cols]
    database[primary_key_cols] = database[primary_key_cols].applymap(str)
    database[primary_key] = database[primary_key_cols].apply(
        lambda x: " ".join(x), axis=1
    )

    new_dataset[primary_key_cols] = new_dataset[primary_key_cols].applymap(str)
    new_dataset[primary_key] = new_dataset[primary_key_cols].apply(
        lambda x: " ".join(x), axis=1
    )

    allupdates = pd.DataFrame(columns=col_order)
    to_append_rows = _append_rows(new_dataset, database, primary_key)
    to_update_rows = _update_rows(new_dataset, database, primary_key)
    if to_delete:
        to_delete_rows = _delete_rows(new_dataset, database, primary_key)
    else:
        to_delete_rows = pd.DataFrame()
    allupdates = pd.concat([allupdates, to_append_rows, to_update_rows], sort=False)
    storedatabase = False
    update_all_file = tempfile.NamedTemporaryFile(dir=SCRIPT_DIR, delete=False)

    with open(update_all_file.name, "w") as updatefile:
        # Must write out the headers in case there are no appends or updates
        updatefile.write(",".join(col_order) + "\n")
        if not allupdates.empty:
            """
            This is done because of pandas typing.
            An integer column with one NA/blank value
            will be cast as a double.
            """
            updatefile.write(
                allupdates[col_order]
                .to_csv(index=False, header=None)
                .replace(".0,", ",")
                .replace(".0\n", "\n")
            )
            storedatabase = True
        if not to_delete_rows.empty:
            updatefile.write(
                to_delete_rows.to_csv(index=False, header=None)
                .replace(".0,", ",")
                .replace(".0\n", "\n")
            )
            storedatabase = True
    if storedatabase:
        syn.store(synapseclient.Table(database_synid, update_all_file.name))
    # Delete the update file
    os.unlink(update_all_file.name)


def checkInt(element):
    """
    Check if an item can become an integer

    Args:
        element: Any variable and type

    Returns:
        boolean: True/False
    """
    try:
        element = float(element)
        return element.is_integer()
    except (ValueError, TypeError):
        return False


def check_col_and_values(
    df, col, possible_values, filename, na_allowed=False, required=False, sep=None
):
    """
    This function checks if the column exists then checks if the values in the
    column have the correct values

    Args:
        df: Input dataframe
        col: Expected column name
        possible_values: list of possible values
        filename: Name of file
        required: If the column is required.  Default is False

    Returns:
        tuple: warning, error
    """
    warning = ""
    error = ""
    have_column = checkColExist(df, col)
    if not have_column:
        if required:
            error = "{filename}: Must have {col} column.\n".format(
                filename=filename, col=col
            )
        else:
            warning = (
                "{filename}: Doesn't have {col} column. "
                "This column will be added\n".format(filename=filename, col=col)
            )
    else:
        if na_allowed:
            check_values = df[col].dropna()
        else:
            check_values = df[col]
        if sep:
            final = []
            for value in check_values:
                final.extend(value.split(sep))
            check_values = pd.Series(final)
        if not check_values.isin(possible_values).all():
            error = "{filename}: Please double check your {col} column.  This column must only be these values: {possible_vals}\n".format(
                filename=filename,
                col=col,
                possible_vals=", ".join(
                    [
                        # This is done because of pandas typing.
                        # An integer column with one NA/blank value
                        # will be cast as a double.
                        str(value).replace(".0", "")
                        for value in possible_values
                    ]
                ),
            )
    return (warning, error)


def extract_oncotree_code_mappings_from_oncotree_json(
    oncotree_json, primary, secondary
):
    oncotree_code_to_info = {}
    data = oncotree_json["children"]
    for node in data:
        # if not node['code']:
        #     sys.stderr.write('Encountered oncotree node without '
        #                       'oncotree code : ' + node + '\n')
        #     continue
        if data[node]["level"] == 1:
            primary = node
            secondary = ""
        elif data[node]["level"] == 2:
            secondary = node
        cancer_type = data[node]["mainType"]
        cancer_type_detailed = data[node]["name"]
        if not cancer_type_detailed:
            cancer_type_detailed = ""
        oncotree_code_to_info[node.upper()] = {
            "CANCER_TYPE": cancer_type,
            "CANCER_TYPE_DETAILED": cancer_type_detailed,
            "ONCOTREE_PRIMARY_NODE": primary,
            "ONCOTREE_SECONDARY_NODE": secondary,
        }

        if len(data[node]["children"]) > 0:
            recurseDict = extract_oncotree_code_mappings_from_oncotree_json(
                data[node], primary, secondary
            )
            oncotree_code_to_info.update(recurseDict)
    return oncotree_code_to_info


def get_oncotree_code_mappings(oncotree_tumortype_api_endpoint_url):
    """
    CREATE ONCOTREE DICTIONARY MAPPING TO PRIMARY, SECONDARY,
    CANCER TYPE, AND CANCER DESCRIPTION
    """
    # oncotree_raw_response = urlopen(oncotree_tumortype_api_endpoint_url).text
    # with requests.get(oncotree_tumortype_api_endpoint_url) as oncotreeUrl:
    oncotreeUrl = retry_get_url(oncotree_tumortype_api_endpoint_url)
    oncotree_raw_response = oncotreeUrl.text
    oncotree_response = json.loads(oncotree_raw_response)
    oncotree_response = oncotree_response["TISSUE"]
    return extract_oncotree_code_mappings_from_oncotree_json(oncotree_response, "", "")


# Get mapping code #Add USE DESCRIPTION sampletypedetailed -> public
def getCODE(mapping, key, useDescription=False):
    if useDescription:
        value = mapping["DESCRIPTION"][mapping["CODE"] == key].values
    else:
        value = mapping["CBIO_LABEL"][mapping["CODE"] == key].values
    if len(value) > 0:
        return value[0]
    else:
        return ""


def getPrimary(code, oncotreeDict, primary):
    if code != "":
        for level in oncotreeDict:
            if sum(oncotreeDict[level] == code) > 0:
                toAdd = primary[oncotreeDict[level] == code].values[0]
                break
            else:
                toAdd = code
    else:
        toAdd = "NOT_ANNOTATED"
    return toAdd


# def create_key():
#   from Crypto.PublicKey import RSA
#   from Crypto import Random
#   from Crypto.Cipher import PKCS1_OAEP
#
#   random_generator = Random.new().read
#   generate public and private keys
#   key = RSA.generate(1024, random_generator)
#   key_cryptor = PKCS1_OAEP.new(key)
#   encrypted = key_cryptor.encrypt(geniePassword)
#   #message to encrypt is in the above line 'encrypt this message'
#   decrypted = key_cryptor.decrypt(encrypted)
#   with open("genie.pem","wb") as geniepem_f:
#       geniepem_f.write(key.exportKey(format='PEM'))


def read_key(pemfile_path):
    """
    Obtain key from pemfile

    Args:
        pemfile_path:  Path to pemfile

    Returns:
        RSA key
    """
    with open(pemfile_path, "r") as pemfile_f:
        key = RSA.importKey(pemfile_f.read())
    return key


def decrypt_message(message, key):
    """
    Decrypt message with a pem key from
    func read_key

    Args:
        message: Encrypted message
        key: read_key returned key

    Returns:
        Decrypted message
    """
    # Use module Crypto.Cipher.PKCS1_OAEP instead
    decryptor = PKCS1_OAEP.new(key)
    decrypted = decryptor.decrypt(ast.literal_eval(str(message)))
    return decrypted.decode("utf-8")


def get_password(pemfile_path):
    """
    Get password using pemfile

    Args:
        pemfile_path: Path to pem file

    Return:
        Password
    """
    if not os.path.exists(pemfile_path):
        raise ValueError(
            "Path to pemFile must be specified if there " "is no cached credentials"
        )
    key = read_key(pemfile_path)
    genie_pass = decrypt_message(os.environ["GENIE_PASS"], key)
    return genie_pass


def synLogin(pemfile_path, debug=False):
    """
    Use pem file to log into synapse if credentials aren't cached

    Args:
        pemfile_path: Path to pem file
        debug: Synapse debug feature.  Defaults to False

    Returns:
        Synapse object logged in
    """
    try:
        syn = synapseclient.Synapse(debug=debug)
        # Get auth token via scheduled job secrets
        if os.getenv("SCHEDULED_JOB_SECRETS") is not None:
            secrets = json.loads(os.getenv("SCHEDULED_JOB_SECRETS"))
            auth_token = secrets["SYNAPSE_AUTH_TOKEN"]
        else:
            auth_token = None
        syn.login(authToken=auth_token)
    except Exception:
        # TODO: deprecate this feature soon
        genie_pass = get_password(pemfile_path)
        syn = synapseclient.Synapse(debug=debug)
        syn.login(os.environ["GENIE_USER"], genie_pass)
    return syn


def get_gdc_data_dictionary(filetype):
    """
    Use the GDC API to get the values allowed for columns of
    different filetypes (ie. disease_type in the case file)

    Args:
        filetype: GDC file type (ie. case, read_group)

    Return:
        json:  Dictionary of allowed columns for the filetype and
               allowed values for those columns
    """
    gdc_dict = retry_get_url(
        "https://api.gdc.cancer.gov/v0/submission/_dictionary/{filetype}".format(
            filetype=filetype
        )
    )
    gdc_response = json.loads(gdc_dict.text)
    return gdc_response


def _create_schema(syn, table_name, parentid, columns=None, annotations=None):
    """Creates Table Schema

    Args:
        syn: Synapse object
        table_name: Name of table
        parentid: Project synapse id
        columns: Columns of Table
        annotations: Dictionary of annotations to add

    Returns:
        Schema
    """
    schema = synapseclient.Schema(
        name=table_name, columns=columns, parent=parentid, annotations=annotations
    )
    new_schema = syn.store(schema)
    return new_schema


def _update_database_mapping(
    syn, database_synid_mappingdf, database_mapping_synid, fileformat, new_tableid
):
    """Updates database to synapse id mapping table

    Args:
        syn: Synapse object
        database_synid_mappingdf: Database to synapse id mapping dataframe
        database_mapping_synid: Database to synapse id table id
        fileformat: File format updated
        new_tableid: New file format table id

    Returns:
        Updated Table object
    """
    fileformat_ind = database_synid_mappingdf["Database"] == fileformat
    # Store in the new database synid
    database_synid_mappingdf["Id"][fileformat_ind] = new_tableid
    # Only update the one row
    to_update_row = database_synid_mappingdf[fileformat_ind]

    syn.store(synapseclient.Table(database_mapping_synid, to_update_row))
    return database_synid_mappingdf


# TODO: deprecate once move function is out of the cli into the
# client master branch
def _move_entity(syn, ent, parentid, name=None):
    """Moves an entity (works like linux mv)

    Args:
        syn: Synapse object
        ent: Synapse Entity
        parentid: Synapse Project id
        name: New Entity name if a new name is desired

    Returns:
        Moved Entity
    """
    ent.parentId = parentid
    if name is not None:
        ent.name = name
    moved_ent = syn.store(ent)
    return moved_ent


def get_dbmapping(syn: Synapse, projectid: str) -> dict:
    """Gets database mapping information

    Args:
        syn: Synapse connection
        projectid: Project id where new data lives

    Returns:
        {'synid': database mapping syn id,
         'df': database mapping pd.DataFrame}

    """
    project_ent = syn.get(projectid)
    dbmapping_synid = project_ent.annotations.get("dbMapping", "")[0]
    database_mapping = syn.tableQuery(f"select * from {dbmapping_synid}")
    database_mappingdf = database_mapping.asDataFrame()
    return {"synid": dbmapping_synid, "df": database_mappingdf}


def create_new_fileformat_table(
    syn: Synapse,
    file_format: str,
    newdb_name: str,
    projectid: str,
    archive_projectid: str,
) -> dict:
    """Creates new database table based on old database table and archives
    old database table

    Args:
        syn: Synapse object
        file_format: File format to update
        newdb_name: Name of new database table
        projectid: Project id where new database should live
        archive_projectid: Project id where old database should be moved

    Returns:
        {"newdb_ent": New database synapseclient.Table,
         "newdb_mappingdf": new databse pd.DataFrame,
         "moved_ent": old database synpaseclient.Table}
    """
    db_info = get_dbmapping(syn, projectid)
    database_mappingdf = db_info["df"]
    dbmapping_synid = db_info["synid"]

    olddb_synid = getDatabaseSynId(
        syn, file_format, databaseToSynIdMappingDf=database_mappingdf
    )
    olddb_ent = syn.get(olddb_synid)
    olddb_columns = list(syn.getTableColumns(olddb_synid))

    newdb_ent = _create_schema(
        syn,
        table_name=newdb_name,
        columns=olddb_columns,
        parentid=projectid,
        annotations=olddb_ent.annotations,
    )

    newdb_mappingdf = _update_database_mapping(
        syn, database_mappingdf, dbmapping_synid, file_format, newdb_ent.id
    )
    # Automatically rename the archived entity with ARCHIVED
    # This will attempt to resolve any issues if the table already exists at
    # location
    new_table_name = f"ARCHIVED {time.time()}-{olddb_ent.name}"
    moved_ent = _move_entity(syn, olddb_ent, archive_projectid, name=new_table_name)
    return {
        "newdb_ent": newdb_ent,
        "newdb_mappingdf": newdb_mappingdf,
        "moved_ent": moved_ent,
    }


def get_database_mapping_config(syn: Synapse, synid: str) -> dict:
    """Gets Synapse database to Table mapping in dict

    Args:
        syn: Synapse connection
        synid: Synapse id of database mapping table

    Returns:
        dict: {'databasename': 'synid'}
    """
    config = syn.tableQuery(f"SELECT * FROM {synid}")
    configdf = config.asDataFrame()
    configdf.index = configdf["Database"]
    config_dict = configdf.to_dict()
    return config_dict["Id"]


def get_genie_config(
    syn: Synapse,
    project_id: str,
) -> dict:
    """Get configurations needed for the GENIE codebase

    Args:
        syn (Synapse): Synapse connection
        project_id (str): Synapse project id

    Returns:
        dict: _description_
    """
    # Get the Synapse Project where data is stored
    # Should have annotations to find the table lookup
    project = syn.get(project_id)

    # Get project GENIE configurations
    database_to_synid_mapping_synid = project.annotations.get("dbMapping", "")
    genie_config = get_database_mapping_config(
        syn=syn, synid=database_to_synid_mapping_synid[0]
    )
    # Fill in GENIE center configurations
    center_mapping_id = genie_config["centerMapping"]
    center_mapping_df = get_syntabledf(
        syn=syn, query_string=f"SELECT * FROM {center_mapping_id} where release is true"
    )
    center_mapping_df.index = center_mapping_df.center
    # Add center configurations including input/staging synapse ids
    genie_config["center_config"] = center_mapping_df.to_dict("index")

    genie_config["ethnicity_mapping"] = "syn7434242"
    genie_config["race_mapping"] = "syn7434236"
    genie_config["sex_mapping"] = "syn7434222"
    genie_config["sampletype_mapping"] = "syn7434273"

    return genie_config


def _get_oncotreelink(syn, genie_config, oncotree_link=None):
    """
    Gets oncotree link unless a link is specified by the user

    Args:
        syn: Synapse object
        databasetosynid_mappingdf: database to synid mapping
        oncotree_link: link to oncotree. Default is None

    Returns:
        oncotree link

    """
    if oncotree_link is None:
        onco_link_ent = syn.get(genie_config["oncotreeLink"])
        oncotree_link = onco_link_ent.externalURL
    return oncotree_link
